import unittest
import time
import scullery.mqtt 
import gc

class VirtualServer(unittest.TestCase):
    def test_loop(self):
        d = [0]
        def p(*x):
            print("Got MQTT response:"+str([x]))
            d[0]+=1
        
        scullery.mqtt.testCrashOnce = True
        c = scullery.mqtt.getConnection("localhost",1883,messageBusName="testBusName")
        
        time.sleep(0.2)
        c.subscribe("test", p)
        
        c.publish("test","someData")
        
        time.sleep(1)
        
        self.assertEqual(d[0], 1)


        #Passive connection, all it does is act as a wrapper around the internal message bus, all real MQTT traffic is through
        #the real connection we made above.
        c2 = scullery.mqtt.getConnection("",messageBusName='testBusName')

        d2=[0]
        def p2(*x):
            print("Got MQTT response2:"+str([x]))
            d2[0]+=1
                
        time.sleep(0.2)
        c2.subscribe("test", p2)
        
        c2.publish("test","someData2")
        
        time.sleep(0.2)
        
        self.assertEqual(d2[0], 1)

        c2.close()
        c.close()

        del c2
        del c
        gc.collect()
        gc.collect()

        #Recreate the connection, all subscribers carry over
        c = scullery.mqtt.getConnection("localhost",1883,messageBusName="testBusName")
        time.sleep(1)
        c.publish("test","someData2")
        time.sleep(0.2)
        self.assertEqual(d2[0], 2)

        #Delete the functions. Scullery only weakly references them, so they should just be gone.
        del p
        del p2
        gc.collect()
        gc.collect()

        c.publish("test","someData2")
        time.sleep(0.2)
        self.assertEqual(d2[0], 2)



# Copyright Daniel Dunn 2019
# This file is part of Scullery.

# Scullery is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.

# Scullery is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Scullery.  If not, see <http://www.gnu.org/licenses/>.

import threading
import weakref
import logging
import time
import uuid
import traceback
import json

from scullery import messagebus, workers, util

try:
    import msgpack
except:
    pass

logger = logging.getLogger("system.mqtt")


connections = {}

lock = threading.RLock()

# {server, topic, qos : weakref(f)}

allSubscriptions = {}

def getWeakrefHandlers(self):
    self = weakref.ref(self)

    def on_connect(client, userdata, flags, rc):
        logger.info("Connected to MQTT server: "+self().server)
        self().onStillConnected()
        # Don't block the network thread too long

        def subscriptionRefresh():
            try:
                with self().lock:
                    for i in allSubscriptions:
                        #Here the bus prefix is our connection ID
                        if i[0] ==self().busPrefix:
                            if allSubscriptions[i]():
                                # Refresh all subscriptions
                                self().connection.subscribe(i[1], i[2])
            except:
                logger.exception("Error subscription refresh")
        workers.do(subscriptionRefresh)

    def on_disconnect(client, userdata, flags, rc):
        logger.info("Disconnected from MQTT server: "+self().server)
        self().onDisconnected()
        logger.info("Disconnected from MQTT server: "+self().server)

    def on_message(client, userdata, msg):
        try:
            s = self()
            # Everything must be fine, because we are getting messages
            messagebus.postMessage(s.busPrefix+"/in/"+msg.topic,
                 msg.payload,
                 )
            s.onStillConnected()
        except Exception:
            print(traceback.format_exc())

    return on_connect, on_disconnect, on_message


#Used to fake a crash fro unit testing purposes
testCrashOnce = False

def makeThread(c,ref):
    def f2():
        global testCrashOnce
        try:
            if testCrashOnce:
                testCrashOnce=False
                raise RuntimeError("Test crash once")
            c.loop_forever(retry_first_connection=True)
        except:
            logger.exception("MQTT Crash")
            ref().onConnectionCrash(traceback.format_exc())
    return f2


class Connection():
    def __init__(self, server, port=1883, password=None, messageBusName=None,*, alertPriority="info", alertAck=True):
        self.server = server
        self.port = port
        self.lock = threading.Lock()
        self.password=password

        self.messageBusName=messageBusName
        self.connection=None

        if not server:
            passive=True
        else:
            passive=False

        server=server or str(uuid.uuid4())
        virtual = server.startswith("__virtual__")


        if passive and (not messageBusName):
            raise ValueError("No server specified. To create a passive connection you must specify an internal messageBusName")
        
     

        self.subscriptions = {}
        logger.info("Creating connection object to: "+self.server)
        import paho.mqtt.client as mqtt

        # When we wrap a function store a weakref to the original here,
        # Pplus the wrapper, so the wrapper doesn't get GCed till
        # The wearkref callback deletes it.
        self.subscribeWrappers = {}

        with lock:
            n = server+":"+str(port)
            if n in connections and connections[n]():
                raise RuntimeError("There is already a connection")
            torm = []
            for i in connections:
                if not connections[i]():
                    torm.append(i)
            for i in torm:
                del connections[i]
            connections[n] = weakref.ref(self)


            self.busPrefix = "/mqtt/"+server+":"+str(port)

            if messageBusName:
                self.busPrefix= "/mqtt/"+messageBusName

            try:

              
                if not virtual and not passive:
                    def out_handler(topic, message,timestamp,annotation):
                        self.connection.publish(topic[len(
                           self.busPrefix +"/out/"):], payload=message, qos=annotation, retain=False)
                else:
                    if not passive:
                        #Virtual loopback server doesn't actually use a real server
                        def out_handler(topic, message,timestamp,annotation):
                            t = topic[len(self.busPrefix+"/out/"):]
                            messagebus.postMessage(self.busPrefix+"/in/"+t, message)

                if not passive:
                    messagebus.subscribe(self.busPrefix+"/out/#", out_handler)
                    self.out_handler = out_handler

                if not virtual and not passive:
                    x = server.split("@")

                    if len(x)==2:
                        host =x[1]
                        user=x[0]
                    elif len(x)==1:
                        host =x[0]
                        user=None
                    else:
                        raise ValueError("More than one @ symbol in server name??")
    

                    self.username = user
                    self.password=password

                    self.configureAlert(alertPriority, alertAck)

                    #Ok so the connection is supposed to do this by itself. Some condition can
                    #Cause the loop to crash and I do not know what!
                    def reconnect():
                        try:
                            if self.connection:
                                self.connection.disconnect()
                        except:
                            pass

                        self.connection = mqtt.Client()
                        # We don't want the connection to stringly reference us, that would interfere with GC
                        on_connect, on_disconnect, on_message = getWeakrefHandlers(
                            self)

                        if self.username:
                            self.username_pw_set(self.username, self.password)

                        self.connection.connect_async(
                            host, port=port, keepalive=60, bind_address="")
                        self.connection.on_connect = on_connect
                        self.connection.on_disconnect = on_disconnect
                        self.connection.on_message = on_message


                        self._thread = threading.Thread(target=makeThread(
                            self.connection,weakref.ref(self)), name=server+":"+str(port), daemon=True)
                        self._thread.start()
                    self.reconnect= reconnect

                    #Actually do the connection.
                    self.reconnect()
                else:
                    self.connection=None
                    self.configureAlert(alertPriority, alertAck)
                    self.onStillConnected()


            except Exception:
                # Attempt cleanup
                try:
                    self.connection.disconnect()
                    time.sleep(2)
                except Exception:
                    pass
                try:
                    del connections[server+":"+str(port)]
                except Exception:
                    pass
                raise

    def onConnectionCrash(self,tb):
        print(traceback.format_exc())
        self.reconnect()

    def onStillConnected(self):
        pass

    def onDisconnected(self):
        pass


    def close(self):
        # Attempt cleanup
        try:
            self.connection.disconnect()
        except Exception:
            pass
        try:
            del connections[server+":"+str(port)]
        except Exception:
            pass
        

    def __del__(self):
        if self.connection:
            self.connection.disconnect()

    def unsubscribe(self, topic, function):
       
        with self.lock:

            #Very expensive to search this dict like this, but unsub shouldn't happen much.
            try:
                torm = []
                #Find any subscriptions to the topic for this particular bus prefix and delete them
                for i in allSubscriptions: 
                    if i[0]==self.busPrefix and i[1]==topic and allSubscriptions[i]()==function:
                        torm.append(i)
                for i in torm:
                    allSubscriptions.pop(i)
            except KeyError:
                print(traceback.format_exc())
                pass

            torm = []
            for i in self.subscribeWrappers:
                x = self.subscribeWrappers[i]
                if x[2] == topic and (x[0]() == function or x[0]() is None):
                    messagebus.unsubscribe(x[3], x[1])
                    torm.append(i)
            for i in torm:
                try:
                    del self.subscribeWrappers[i]
                except KeyError:
                    pass

            for i in self.subscribeWrappers:
                x = self.subscribeWrappers[i]
                if x[2] == topic:
                    return

            # We could not find even a single subscriber function
            # So we unsubscribe at the MQTT level
            logging.debug("MQTT Unsubscribe from "+topic+" at "+self.server)
            if self.connection:
                self.connection.unsubscribe(topic)

    def configureAlert(self, *a):
        pass

    def subscribe(self, topic, function, qos=2, encoding="json"):
        if self.connection:
            self.connection.subscribe(topic, qos)
       
        x = str(uuid.uuid4())
        

        def handleDel(*a):
            del self.subscribeWrappers[x]
            # We're really just using the "check if there's no subscribers"
            # Part of the function
            self.unsubscribe(topic, None)
        function = util.universal_weakref(function, handleDel)

        #This is our master list of who is subscribed where. It is used for reconnection.
        #It is also used so that subscriptions can persist beyond any given connection object!!!
        with self.lock:
            allSubscriptions[self.busPrefix,topic,qos]= function


        # Connection.subscribe was blocking forever.
        # Use a different thread, to hopefully avoid deadlocks
        def backgroundSubscribeTask():
            with self.lock:
                self.connection.subscribe(topic, qos)
        

        if encoding == 'json':
            def wrapper(t, m):
                # Get rid of the extra kaithem framing part of the topic
                t = t[len(self.busPrefix+"/in/"):]
                function()(t, json.loads(m))

        elif encoding == 'msgpack':
            def wrapper(t, m):
                # Get rid of the extra kaithem framing part of the topic
                t = t[len(self.busPrefix+"/in/"):]
                function()(t, msgpack.unpackb(m,raw=False))

        elif encoding == 'utf8':
            def wrapper(t, m):
                # Get rid of the extra kaithem framing part of the topic
                t = t[len(self.busPrefix+"/in/"):]
                function()(t, m.decode("utf8"))

        elif encoding == 'raw':
            def wrapper(t, m):
                # Get rid of the extra kaithem framing part of the topic
                t = t[len(self.busPrefix+"/in/"):]
                function()(t, m)
        else:
            raise ValueError("Invalid encoding: "+encoding)

        # Correctly call message bus error handlers
        wrapper.messagebusWrapperFor = function

        internalTopic = self.busPrefix+"/in/"+topic

        # Extra data is mostly used for unsubscription
        self.subscribeWrappers[x] = (function, wrapper, topic, internalTopic)

        logging.debug("MQTT subscribe to "+topic+" at "+self.server)
        # Ref to f exists as long as the original does because it's kept in subscribeWrappers
        messagebus.subscribe(internalTopic, wrapper)

        #Important we do this *after*, because the server might auto-send us something that could be important for.
        #Only the first function will get it, but whatever, at least we can see certain things when manually subscribing, for test purposes.
        #Or should we do it before, to resolve that bit of ambiguity and never get the message???
        if self.connection:
            workers.do(backgroundSubscribeTask)

    def publish(self, topic, message, qos=2, encoding="json"):
        if encoding == 'json':
            message = json.dumps(message)
        elif encoding == 'msgpack':
            message = msgpack.packb(message,use_bin_type=True)
        elif encoding == 'utf8':
            message = message.encode("utf8")
        elif encoding == 'raw':
            pass
        else:
            raise ValueError("Invalid encoding!")
        messagebus.postMessage(self.busPrefix+"/out/"+topic, message, annotation=qos)


def getConnection(server, port=1883, password=None,messageBusName=None,*, alertPriority="info", alertAck=True):
    # Kaithem is gonna monkeypatch kaithem with one that has better
    # logging
    global Connection

    #Blank password means no password
    password=password or None
    with lock:
        x = None
        
        if server+":"+str(port) in connections:
            x = connections[server+":"+str(port)]()

        if x:
            if not messageBusName==x.messageBusName:
                raise RuntimeError("Connection already exists, but with a different message bus name. ")

            if x.password or password:
                if not x.password== password:
                    raise ValueError("There is already a connection to the same host and user, but with a different password.")

            x.configureAlert(alertPriority, alertAck)
            return x

        return Connection(server, port, password=password,alertAck=True, alertPriority="info",messageBusName=messageBusName)



import sys,os,weakref,types
#Credit to Jay of stack overflow for this function
def which(program):
    "Check if a program is installed like you would do with UNIX's which command."

    #Because in windows, the actual executable name has .exe while the command name does not.
    if sys.platform == "win32" and not program.endswith(".exe"):
        program += ".exe"

    #Find out if path represents a file that the current user can execute.
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    #If the input was a direct path to an executable, return it
    if fpath:
        if is_exe(program):
            return program

    #Else search the path for the file.
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    #If we got this far in execution, we assume the file is not there and return None
    return None

def universal_weakref(f,cb=None):
    if isinstance(f,types.MethodType):
        return weakref.WeakMethod(f,cb)
    else:
        return weakref.ref(f,cb)



"""
================================================================================

    pyFluidSynth

    Python bindings for FluidSynth

    Copyright 2008-2019, Nathan Whitehead <nwhitehe@gmail.com> and others


    Released under the LGPL

    This module contains python bindings for FluidSynth.  FluidSynth is a
    software synthesizer for generating music.  It works like a MIDI
    synthesizer.  You load patches, set parameters, then send NOTEON and
    NOTEOFF events to play notes.  Instruments are defined in SoundFonts,
    generally files with the extension SF2.  FluidSynth can either be used
    to play audio itself, or you can call a function that returns chunks
    of audio data and output the data to the soundcard yourself.
    FluidSynth works on all major platforms, so pyFluidSynth should also.

================================================================================

"""

# Standard library modules
from ctypes import (CDLL, CFUNCTYPE, POINTER, Structure, byref, c_char, c_char_p, c_double,
                    c_float, c_int, c_short, c_size_t, c_uint, c_void_p, create_string_buffer)
from ctypes.util import find_library
import os

# Third-party modules
from six import binary_type, iteritems, text_type

# Constants

# Bump this up when changing the interface for users
api_version = '2.0'
# Used to encode / decode strings passed to / from C functions
DEFAULT_ENCODING = 'utf-8'
# Function call result
FLUID_OK = 0
FLUID_FAILED = -1
# Settings types
FLUID_NO_TYPE = -1  # Undefined type
FLUID_NUM_TYPE = 0  # Numeric (double)
FLUID_INT_TYPE = 1  # Integer
FLUID_STR_TYPE = 2  # String
FLUID_SET_TYPE = 3  # Set of values.
# MIDI router rule types
FLUID_MIDI_ROUTER_RULE_NOTE = 0              # MIDI note rule
FLUID_MIDI_ROUTER_RULE_CC = 1                # MIDI controller rule
FLUID_MIDI_ROUTER_RULE_PROG_CHANGE = 2       # MIDI program change rule
FLUID_MIDI_ROUTER_RULE_PITCH_BEND = 3        # MIDI pitch bend rule
FLUID_MIDI_ROUTER_RULE_CHANNEL_PRESSURE = 4  # MIDI channel pressure rule
FLUID_MIDI_ROUTER_RULE_KEY_PRESSURE = 5      # MIDI key pressure rule
# MIDI player state
FLUID_PLAYER_READY = 0  # Player is ready
FLUID_PLAYER_PLAYING = 1  # Player is currently playing
FLUID_PLAYER_DONE = 2  # Player is finished playing
# Driver names
AUDIO_DRIVER_NAMES = ("alsa, coreaudio, dart, dsound, file, jack, oss, portaudio, pulseaudio, "
                      "sdl2, sndman, waveout").split(", ")
MIDI_DRIVER_NAMES = "alsa_raw, alsa_seq, coremidi, jack, midishare, oss, winmidi".split(", ")
AUDIO_FILE_TYPES = ("aiff, au, auto, avr, caf, flac, htk, iff, mat, oga, paf, pvf, raw, sd2, sds, "
                    "sf, voc, w64, wav, xi").split(", ")

# A short circuited or expression to find the FluidSynth library
# (mostly needed for Windows distributions of libfluidsynth supplied with QSynth)
lib = (find_library('fluidsynth') or
       find_library('libfluidsynth') or
       find_library('libfluidsynth-2') or
       find_library('libfluidsynth-1'))

if lib is None:
    raise ImportError("Couldn't find the FluidSynth library.")

# Dynamically link the FluidSynth library
_fl = CDLL(lib)


# Helper function for declaring function prototypes
def cfunc(name, result, *args):
    """Build and apply a ctypes prototype complete with parameter flags."""
    atypes = []
    aflags = []
    for arg in args:
        atypes.append(arg[1])
        aflags.append((arg[2], arg[0]) + arg[3:])
    return (CFUNCTYPE(result, *atypes)((name, _fl), tuple(aflags)))


# Function prototypes for C versions of functions

# Fluid settings
new_fluid_settings = cfunc(
    'new_fluid_settings',
    c_void_p)
fluid_settings_get_type = cfunc(
    'fluid_settings_get_type',
    c_int,
    ('settings', c_void_p, 1),
    ('name', c_char_p, 1))
fluid_settings_copystr = cfunc(
    'fluid_settings_copystr',
    c_int,
    ('settings', c_void_p, 1),
    ('name', c_char_p, 1),
    ('str', c_char_p, 1),
    ('len', c_int, 1))
fluid_settings_getnum = cfunc(
    'fluid_settings_getnum',
    c_int,
    ('settings', c_void_p, 1),
    ('name', c_char_p, 1),
    ('val', POINTER(c_double), 1))
fluid_settings_getint = cfunc(
    'fluid_settings_getint',
    c_int,
    ('settings', c_void_p, 1),
    ('name', c_char_p, 1),
    ('val', POINTER(c_int), 1))
fluid_settings_setstr = cfunc(
    'fluid_settings_setstr',
    c_int,
    ('settings', c_void_p, 1),
    ('name', c_char_p, 1),
    ('str', c_char_p, 1))
fluid_settings_setnum = cfunc(
    'fluid_settings_setnum',
    c_int,
    ('settings', c_void_p, 1),
    ('name', c_char_p, 1),
    ('val', c_double, 1))
fluid_settings_setint = cfunc(
    'fluid_settings_setint',
    c_int,
    ('settings', c_void_p, 1),
    ('name', c_char_p, 1),
    ('val', c_int, 1))
delete_fluid_settings = cfunc(
    'delete_fluid_settings',
    None,
    ('settings', c_void_p, 1))

# Fluid synth
new_fluid_synth = cfunc(
    'new_fluid_synth',
    c_void_p,
    ('settings', c_void_p, 1))
delete_fluid_synth = cfunc(
    'delete_fluid_synth',
    None,
    ('synth', c_void_p, 1))
# Soundfont handling
fluid_synth_sfload = cfunc(
    'fluid_synth_sfload',
    c_int,
    ('synth', c_void_p, 1),
    ('filename', c_char_p, 1),
    ('update_midi_presets', c_int, 1))
fluid_synth_sfunload = cfunc(
    'fluid_synth_sfunload',
    c_int,
    ('synth', c_void_p, 1),
    ('sfid', c_int, 1),
    ('update_midi_presets', c_int, 1))
fluid_synth_get_sfont_by_id = cfunc(
    'fluid_synth_get_sfont_by_id',
    c_void_p,
    ('synth', c_void_p, 1),
    ('id', c_int, 1))
fluid_synth_sfont_select = cfunc(
    'fluid_synth_sfont_select',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1),
    ('sfid', c_int, 1))
# MIDI program handling
fluid_synth_program_select = cfunc(
    'fluid_synth_program_select',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1),
    ('sfid', c_int, 1),
    ('bank', c_int, 1),
    ('preset', c_int, 1))
fluid_synth_unset_program = cfunc(
    'fluid_synth_unset_program',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1))
fluid_synth_get_program = cfunc(
    'fluid_synth_get_program',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1),
    ('sfont_id', POINTER(c_int), 1),
    ('bank_num', POINTER(c_int), 1),
    ('preset_num', POINTER(c_int), 1))
# MIDI events
fluid_synth_noteon = cfunc(
    'fluid_synth_noteon',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1),
    ('key', c_int, 1),
    ('vel', c_int, 1))
fluid_synth_noteoff = cfunc(
    'fluid_synth_noteoff',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1),
    ('key', c_int, 1))
fluid_synth_pitch_bend = cfunc(
    'fluid_synth_pitch_bend',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1),
    ('val', c_int, 1))
fluid_synth_cc = cfunc(
    'fluid_synth_cc',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1),
    ('ctrl', c_int, 1),
    ('val', c_int, 1))
fluid_synth_get_cc = cfunc(
    'fluid_synth_get_cc',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1),
    ('num', c_int, 1),
    ('pval', POINTER(c_int), 1))
fluid_synth_program_change = cfunc(
    'fluid_synth_program_change',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1),
    ('prg', c_int, 1))
fluid_synth_bank_select = cfunc(
    'fluid_synth_bank_select',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1),
    ('bank', c_int, 1))
fluid_synth_all_notes_off = cfunc(
    'fluid_synth_all_notes_off',
    c_int,
    ('synth', c_void_p, 1),
    ('chan', c_int, 1))
# Reset functions
fluid_synth_program_reset = cfunc(
    'fluid_synth_program_reset',
    c_int,
    ('synth', c_void_p, 1))
fluid_synth_system_reset = cfunc(
    'fluid_synth_system_reset',
    c_int,
    ('synth', c_void_p, 1))
# Misc
fluid_synth_write_s16 = cfunc(
    'fluid_synth_write_s16',
    c_void_p,
    ('synth', c_void_p, 1),
    ('len', c_int, 1),
    ('lbuf', c_void_p, 1),
    ('loff', c_int, 1),
    ('lincr', c_int, 1),
    ('rbuf', c_void_p, 1),
    ('roff', c_int, 1),
    ('rincr', c_int, 1))
fluid_synth_handle_midi_event = cfunc(
    'fluid_synth_handle_midi_event',
    c_int,
    ('data', c_void_p, 1),
    ('event', c_void_p, 1))

# Reverb
fluid_synth_get_reverb_roomsize = cfunc(
    'fluid_synth_get_reverb_roomsize',
    c_double,
    ('synth', c_void_p, 1))
fluid_synth_get_reverb_damp = cfunc(
    'fluid_synth_get_reverb_damp',
    c_double,
    ('synth', c_void_p, 1))
fluid_synth_get_reverb_level = cfunc(
    'fluid_synth_get_reverb_level',
    c_double,
    ('synth', c_void_p, 1))
fluid_synth_get_reverb_width = cfunc(
    'fluid_synth_get_reverb_width',
    c_double,
    ('synth', c_void_p, 1))

try:
    fluid_synth_set_reverb = cfunc(
        'fluid_synth_set_reverb',
        c_int,
        ('synth', c_void_p, 1),
        ('roomsize', c_double, 1),
        ('damping', c_double, 1),
        ('width', c_double, 1),
        ('level', c_double, 1))
except AttributeError:
    fluid_synth_set_reverb_full = cfunc(
        'fluid_synth_set_reverb_full',
        c_int,
        ('synth', c_void_p, 1),
        ('set', c_int, 1),
        ('roomsize', c_double, 1),
        ('damping', c_double, 1),
        ('width', c_double, 1),
        ('level', c_double, 1))
    fluid_synth_set_reverb = None

# Handle fluidsynth >= 2.0 API changes
try:
    fluid_synth_set_reverb_roomsize = cfunc(
        'fluid_synth_set_reverb_roomsize',
        c_int,
        ('synth', c_void_p, 1),
        ('roomsize', c_double, 1))
    fluid_synth_set_reverb_damp = cfunc(
        'fluid_synth_set_reverb_damp',
        c_int,
        ('synth', c_void_p, 1),
        ('damping', c_double, 1))
    fluid_synth_set_reverb_level = cfunc(
        'fluid_synth_set_reverb_level',
        c_int,
        ('synth', c_void_p, 1),
        ('level', c_double, 1))
    fluid_synth_set_reverb_width = cfunc(
        'fluid_synth_set_reverb_width',
        c_int,
        ('synth', c_void_p, 1),
        ('width', c_double, 1))
except AttributeError:
    fluid_synth_set_reverb_roomsize = None
    fluid_synth_set_reverb_damp = None
    fluid_synth_set_reverb_level = None
    fluid_synth_set_reverb_width = None

# Chorus
fluid_synth_get_chorus_nr = cfunc(
    'fluid_synth_get_chorus_nr',
    c_int,
    ('synth', c_void_p, 1))
fluid_synth_get_chorus_level = cfunc(
    'fluid_synth_get_chorus_level',
    c_double,
    ('synth', c_void_p, 1))
fluid_synth_get_chorus_type = cfunc(
    'fluid_synth_get_chorus_type',
    c_int,
    ('synth', c_void_p, 1))

try:
    fluid_synth_set_chorus = cfunc(
        'fluid_synth_set_chorus',
        c_int,
        ('synth', c_void_p, 1),
        ('nr', c_int, 1),
        ('level', c_double, 1),
        ('speed', c_double, 1),
        ('depth_ms', c_double, 1),
        ('type', c_int, 1))
except AttributeError:
    fluid_synth_set_chorus_full = cfunc(
        'fluid_synth_set_chorus_full',
        c_int,
        ('synth', c_void_p, 1),
        ('set', c_int, 1),
        ('nr', c_int, 1),
        ('level', c_double, 1),
        ('speed', c_double, 1),
        ('depth_ms', c_double, 1),
        ('type', c_int, 1))
    fluid_synth_set_chorus = None

# Handle fluidsynth >= 2.0 API changes
try:
    fluid_synth_set_chorus_nr = cfunc(
        'fluid_synth_set_chorus_nr',
        c_int,
        ('synth', c_void_p, 1),
        ('nr', c_int, 1))
    fluid_synth_set_chorus_level = cfunc(
        'fluid_synth_set_chorus_level',
        c_int,
        ('synth', c_void_p, 1),
        ('level', c_double, 1))
    fluid_synth_set_chorus_type = cfunc(
        'fluid_synth_set_chorus_type',
        c_int,
        ('synth', c_void_p, 1),
        ('type', c_int, 1))
    fluid_synth_set_chorus_speed = cfunc(
        'fluid_synth_set_chorus_speed',
        c_int,
        ('synth', c_void_p, 1),
        ('speed', c_double, 1))
    fluid_synth_set_chorus_depth = cfunc(
        'fluid_synth_set_chorus_depth',
        c_int,
        ('synth', c_void_p, 1),
        ('depth', c_double, 1))
    fluid_synth_get_chorus_speed = fluid_synth_get_chorus_speed_Hz = cfunc(
        'fluid_synth_get_chorus_speed',
        c_double,
        ('synth', c_void_p, 1))
    fluid_synth_get_chorus_depth = fluid_synth_get_chorus_depth_ms = cfunc(
        'fluid_synth_get_chorus_depth',
        c_double,
        ('synth', c_void_p, 1))
except AttributeError:
    fluid_synth_get_chorus_speed = fluid_synth_get_chorus_speed_Hz = cfunc(
        'fluid_synth_get_chorus_speed_Hz',
        c_double,
        ('synth', c_void_p, 1))
    fluid_synth_get_chorus_depth = fluid_synth_get_chorus_depth_ms = cfunc(
        'fluid_synth_get_chorus_depth_ms',
        c_double,
        ('synth', c_void_p, 1))

try:
    fluid_synth_set_midi_router = cfunc(
        'fluid_synth_set_midi_router',
        None,
        ('synth', c_void_p, 1),
        ('router', c_void_p, 1))
except AttributeError:
    # fluidsynth >= 2.0
    fluid_synth_set_midi_router = None

try:
    class fluid_synth_channel_info_t(Structure):
        _fields_ = [
            ('assigned', c_int),
            ('sfont_id', c_int),
            ('bank', c_int),
            ('program', c_int),
            ('name', c_char * 32),
            ('reserved', c_char * 32)]

    fluid_synth_get_channel_info = cfunc(
        'fluid_synth_get_channel_info',
        c_int,
        ('synth', c_void_p, 1),
        ('chan', c_int, 1),
        ('info', POINTER(fluid_synth_channel_info_t), 1))
except AttributeError:
    fluid_synth_get_channel_info = None
    del fluid_synth_channel_info_t

# Fluid audio driver
new_fluid_audio_driver = cfunc(
    'new_fluid_audio_driver',
    c_void_p,
    ('settings', c_void_p, 1),
    ('synth', c_void_p, 1))
delete_fluid_audio_driver = cfunc(
    'delete_fluid_audio_driver',
    None,
    ('driver', c_void_p, 1))

# Fluid MIDI driver
new_fluid_midi_driver = cfunc(
    'new_fluid_midi_driver',
    c_void_p,
    ('settings', c_void_p, 1),
    ('handler', CFUNCTYPE(c_int, c_void_p, c_void_p), 1),
    ('event_handler_data', c_void_p, 1))
delete_fluid_midi_driver = cfunc(
    'delete_fluid_midi_driver',
    None,
    ('driver', c_void_p, 1))


# Fluid MIDI router
class fluid_midi_router_t(Structure):
    _fields_ = [
        ('synth', c_void_p),
        ('rules_mutex', c_void_p),
        ('rules', c_void_p * 6),
        ('free_rules', c_void_p),
        ('event_handler', c_void_p),
        ('event_handler_data', c_void_p),
        ('nr_midi_channels', c_int),
        ('cmd_rule', c_void_p),
        ('cmd_rule_type', POINTER(c_int))]


new_fluid_midi_router = cfunc(
    'new_fluid_midi_router',
    POINTER(fluid_midi_router_t),
    ('settings', c_void_p, 1),
    ('handler', CFUNCTYPE(c_int, c_void_p, c_void_p), 1),
    ('event_handler_data', c_void_p, 1))
fluid_midi_router_handle_midi_event = cfunc(
    'fluid_midi_router_handle_midi_event',
    c_int,
    ('data', c_void_p, 1),
    ('event', c_void_p, 1))
fluid_midi_dump_prerouter = cfunc(
    'fluid_midi_dump_prerouter',
    c_int,
    ('data', c_void_p, 1),
    ('event', c_void_p, 1))
fluid_midi_dump_postrouter = cfunc(
    'fluid_midi_dump_postrouter',
    c_int,
    ('data', c_void_p, 1),
    ('event', c_void_p, 1))
fluid_midi_router_clear_rules = cfunc(
    'fluid_midi_router_clear_rules',
    c_int,
    ('router', POINTER(fluid_midi_router_t), 1))
fluid_midi_router_set_default_rules = cfunc(
    'fluid_midi_router_set_default_rules',
    c_int,
    ('router', POINTER(fluid_midi_router_t), 1))
fluid_midi_router_add_rule = cfunc(
    'fluid_midi_router_add_rule',
    c_int,
    ('router', POINTER(fluid_midi_router_t), 1),
    ('rule', c_void_p, 1),
    ('type', c_int, 1))
delete_fluid_midi_router = cfunc(
    'delete_fluid_midi_router',
    c_int,
    ('router', POINTER(fluid_midi_router_t), 1))

# Fluid MIDI router rules
new_fluid_midi_router_rule = cfunc(
    'new_fluid_midi_router_rule',
    c_void_p)
fluid_midi_router_rule_set_chan = cfunc(
    'fluid_midi_router_rule_set_chan',
    None,
    ('rule', c_void_p, 1),
    ('min', c_int, 1),
    ('max', c_int, 1),
    ('mul', c_float, 1),
    ('add', c_int, 1))
fluid_midi_router_rule_set_param1 = cfunc(
    'fluid_midi_router_rule_set_param1',
    None,
    ('rule', c_void_p, 1),
    ('min', c_int, 1),
    ('max', c_int, 1),
    ('mul', c_float, 1),
    ('add', c_int, 1))
fluid_midi_router_rule_set_param2 = cfunc(
    'fluid_midi_router_rule_set_param2',
    None,
    ('rule', c_void_p, 1),
    ('min', c_int, 1),
    ('max', c_int, 1),
    ('mul', c_float, 1),
    ('add', c_int, 1))
delete_fluid_midi_router_rule = cfunc(
    'delete_fluid_midi_router_rule',
    c_int,
    ('rule', c_void_p, 1))

# Command handler
try:
    new_fluid_cmd_handler = cfunc(
        'new_fluid_cmd_handler',
        c_void_p,
        ('synth', c_void_p, 1),
        ('router', c_void_p, 1))
except AttributeError:
    new_fluid_cmd_handler = cfunc(
        'new_fluid_cmd_handler',
        c_void_p,
        ('synth', c_void_p, 1))

delete_fluid_cmd_handler = cfunc(
        'delete_fluid_cmd_handler',
        c_void_p,
        ('handler', c_void_p, 1))

# Preset handling
try:
    fluid_preset_get_name = cfunc(
        'fluid_preset_get_name',
        c_char_p,
        ('preset', c_void_p, 1))
    fluid_sfont_get_preset = cfunc(
        'fluid_sfont_get_preset',
        c_void_p,
        ('sfont', c_void_p, 1),
        ('banknum', c_int, 1),
        ('prognum', c_int, 1))
except AttributeError:
    fluid_preset_get_name = None
    fluid_sfont_get_preset = None

# Fluid file renderer
new_fluid_file_renderer = cfunc(
    'new_fluid_file_renderer',
    c_void_p,
    ('synth', c_void_p, 1))
fluid_file_renderer_process_block = cfunc(
    'fluid_file_renderer_process_block',
    c_int,
    ('dev', c_void_p, 1))
fluid_file_set_encoding_quality = cfunc(
    'fluid_file_set_encoding_quality',
    c_int,
    ('dev', c_void_p, 1),
    ('q', c_double, 1))
delete_fluid_file_renderer = cfunc(
    'delete_fluid_file_renderer',
    c_void_p,
    ('dev', c_void_p, 1))

# Fluid MIDI player
new_fluid_player = cfunc(
    'new_fluid_player',
    c_void_p,
    ('synth', c_void_p, 1))
fluid_player_add = cfunc(
    'fluid_player_add',
    c_int,
    ('player', c_void_p, 1),
    ('midifile', c_char_p, 1))
fluid_player_add_mem = cfunc(
    'fluid_player_add_mem',
    c_int,
    ('player', c_void_p, 1),
    ('buffer', c_void_p, 1),
    ('len', c_size_t, 1))
fluid_player_get_bpm = cfunc(
    'fluid_player_get_bpm',
    c_int,
    ('player', c_void_p, 1))
fluid_player_get_midi_tempo = cfunc(
    'fluid_player_get_midi_tempo',
    c_int,
    ('player', c_void_p, 1))
fluid_player_get_status = cfunc(
    'fluid_player_get_status',
    c_int,
    ('player', c_void_p, 1))
fluid_player_get_current_tick = cfunc(
    'fluid_player_get_current_tick',
    c_int,
    ('player', c_void_p, 1))
fluid_player_get_total_ticks = cfunc(
    'fluid_player_get_total_ticks',
    c_int,
    ('player', c_void_p, 1))
fluid_player_join = cfunc(
    'fluid_player_join',
    c_int,
    ('player', c_void_p, 1))
fluid_player_play = cfunc(
    'fluid_player_play',
    c_int,
    ('player', c_void_p, 1))

#FS 1.0 doesn't have this.
try:
    fluid_player_seek = cfunc(
        'fluid_player_seek',
        c_int,
        ('player', c_void_p, 1),
        ('ticks', c_int, 1))
except AttributeError:
    pass

fluid_player_set_loop = cfunc(
    'fluid_player_set_loop',
    c_int,
    ('player', c_void_p, 1),
    ('loop', c_int, 1))
fluid_player_set_bpm = cfunc(
    'fluid_player_set_bpm',
    c_int,
    ('player', c_void_p, 1),
    ('bpm', c_int, 1))
fluid_player_set_midi_tempo = cfunc(
    'fluid_player_set_midi_tempo',
    c_int,
    ('player', c_void_p, 1),
    ('tempo', c_int, 1))
fluid_player_stop = cfunc(
    'fluid_player_stop',
    c_int,
    ('player', c_void_p, 1))
delete_fluid_player = cfunc(
    'delete_fluid_player',
    c_void_p,
    ('player', c_void_p, 1))

# Fluid sequencer
new_fluid_sequencer2 = cfunc(
    'new_fluid_sequencer2',
    c_void_p,
    ('use_system_timer', c_int, 1))
fluid_sequencer_process = cfunc(
    'fluid_sequencer_process',
    None,
    ('seq', c_void_p, 1),
    ('msec', c_uint, 1))
fluid_sequencer_register_fluidsynth = cfunc(
    'fluid_sequencer_register_fluidsynth',
    c_short,
    ('seq', c_void_p, 1),
    ('synth', c_void_p, 1))
fluid_sequencer_register_client = cfunc(
    'fluid_sequencer_register_client',
    c_short,
    ('seq', c_void_p, 1),
    ('name', c_char_p, 1),
    ('callback', CFUNCTYPE(None, c_uint, c_void_p, c_void_p, c_void_p), 1),
    ('data', c_void_p, 1))
fluid_sequencer_get_tick = cfunc(
    'fluid_sequencer_get_tick',
    c_uint,
    ('seq', c_void_p, 1))
fluid_sequencer_set_time_scale = cfunc(
    'fluid_sequencer_set_time_scale',
    None,
    ('seq', c_void_p, 1),
    ('scale', c_double, 1))
fluid_sequencer_get_time_scale = cfunc(
    'fluid_sequencer_get_time_scale',
    c_double,
    ('seq', c_void_p, 1))
fluid_sequencer_send_at = cfunc(
    'fluid_sequencer_send_at',
    c_int,
    ('seq', c_void_p, 1),
    ('evt', c_void_p, 1),
    ('time', c_uint, 1),
    ('absolute', c_int, 1))
delete_fluid_sequencer = cfunc(
    'delete_fluid_sequencer',
    None,
    ('seq', c_void_p, 1))

# Fluid events
new_fluid_event = cfunc(
    'new_fluid_event',
    c_void_p)
fluid_event_set_source = cfunc(
    'fluid_event_set_source',
    None,
    ('evt', c_void_p, 1),
    ('src', c_void_p, 1))
fluid_event_set_dest = cfunc(
    'fluid_event_set_dest',
    None,
    ('evt', c_void_p, 1),
    ('dest', c_void_p, 1))
fluid_event_timer = cfunc(
    'fluid_event_timer',
    None,
    ('evt', c_void_p, 1),
    ('data', c_void_p, 1))
fluid_event_note = cfunc(
    'fluid_event_note',
    None,
    ('evt', c_void_p, 1),
    ('channel', c_int, 1),
    ('key', c_short, 1),
    ('vel', c_short, 1),
    ('duration', c_uint, 1))
fluid_event_noteon = cfunc(
    'fluid_event_noteon',
    None,
    ('evt', c_void_p, 1),
    ('channel', c_int, 1),
    ('key', c_short, 1),
    ('vel', c_short, 1))
fluid_event_noteoff = cfunc(
    'fluid_event_noteoff',
    None,
    ('evt', c_void_p, 1),
    ('channel', c_int, 1),
    ('key', c_short, 1))
delete_fluid_event = cfunc(
    'delete_fluid_event',
    None,
    ('evt', c_void_p, 1))


# (Internal) Helper functions

def _d(s, encoding=DEFAULT_ENCODING):
    if isinstance(s, binary_type) and encoding:
        return s.decode(encoding)
    return s


def _e(s, encoding=DEFAULT_ENCODING):
    if isinstance(s, text_type) and encoding:
        return s.encode(encoding)
    return s


# Convenience functions

def fluid_synth_write_s16_stereo(synth, nframes):
    """Return generated samples in stereo 16-bit format.

    :param synth: an instance of class Synth
    :param nframes: number of sample frames to generate
    :type nframes: ``int``
    :return: one-dimenional NumPy array of interleaved samples
    :rtype: ``np.array(..., dtype=numpy.int16)``

    """
    import numpy
    buf = create_string_buffer(nframes * 4)
    fluid_synth_write_s16(synth, nframes, buf, 0, 2, buf, 1, 2)
    return numpy.frombuffer(buf[:], dtype=numpy.int16)


def raw_audio_string(data):
    """Return a string of bytes to send to soundcard.

    Input is a numpy array of samples. Default output format is 16-bit signed
    (other formats not currently supported).

    """
    import numpy
    return (data.astype(numpy.int16)).tostring()


# Object-oriented interface, simplifies access to functions

class RouterRule:
    """Represents a FluidSynth MIDI router rule."""

    def __init__(self, type=FLUID_MIDI_ROUTER_RULE_NOTE, chan=None, param1=None, param2=None):
        self.type = type
        self.rule = new_fluid_midi_router_rule()

        if isinstance(chan, dict):
            self.set_chan(**chan)
        elif chan is not None:
            self.set_chan(*chan)

        if isinstance(param1, dict):
            self.set_param1(**param1)
        elif param1 is not None:
            self.set_param1(*param1)

        if isinstance(param2, dict):
            self.set_param2(**param2)
        elif param2 is not None:
            self.set_param2(*param2)

    def delete(self):
        if self.rule:
            delete_fluid_midi_router_rule(self.rule)
            self.rule = None

    def set_chan(self, min=0, max=15, mul=1.0, add=0):
        if self.rule:
            fluid_midi_router_rule_set_chan(self.rule, min, max, mul, add)

    def set_param1(self, min=0, max=127, mul=1.0, add=0):
        if self.rule:
            fluid_midi_router_rule_set_param1(self.rule, min, max, mul, add)

    def set_param2(self, min=0, max=127, mul=1.0, add=0):
        if self.rule:
            fluid_midi_router_rule_set_param2(self.rule, min, max, mul, add)


class BasePlayer(object):
    """Interface for the FluidSynth internal MIDI player."""

    def __init__(self, synth):
        """Initialize Player instance.

        :param synth: an instance of class Synth

        """
        self.synth = synth
        self.player = new_fluid_player(self.synth.synth)

    def delete(self):
        delete_fluid_player(self.player)

    def add(self, filename):
        """Add Standard MIDI File to the playlist.

        :param filename: SMF name / path
        :type filename: ``str``

        """
        return fluid_player_add(self.player, _e(filename))

    def add_mem(self, data):
        """Add SMF data to the playlist.

        :param data: SMF MIDI data
        :type data: ``bytes`` or ``bytearray``

        """
        return fluid_player_add_mem(self.player, data, len(data))

    def play(self):
        """Start playing."""
        return fluid_player_play(self.player)

    def join(self):
        """Wait until player is finished playing."""
        return fluid_player_join(self.player)

    def stop(self):
        """Stop playing."""
        return fluid_player_stop(self.player)

    @property
    def bpm(self):
        """Return current player tempo in BPM.

        :return: tempo in beats (quarter notes) per minute

        """
        return fluid_player_get_bpm(self.player)

    @bpm.setter
    def bpm(self, beats_per_minute):
        """Set player tempo in BPM.

        :param: beats_per_minute: tempo in beats (quarter notes) per minute
        :type beats_per_minute: ``int``

        """
        return fluid_player_set_bpm(self.player, beats_per_minute)

    @property
    def current_tick(self):
        """Return current player position.

        :return: position in MIDI ticks
        :rtype: ``int``

        """
        return fluid_player_get_current_tick(self.player)

    @current_tick.setter
    def current_tick(self, ticks):
        """Seek to given player position.

        :param ticks: position in MIDI ticks
        :type ticks: ``int``

        """
        return fluid_player_seek(self.player, ticks)

    @property
    def status(self):
        """Return player status.

        :return: ``FLUID_PLAYER_READY``, ``FLUID_PLAYER_PLAYING``,
            or ``FLUID_PLAYER_DONE`` constant value

        """
        return fluid_player_get_status(self.player)

    @property
    def tempo(self):
        """Return current player tempo.

        :return: tempo in microseconds per beat (quarter note)
        :rtype: ``int``

        """
        return fluid_player_get_midi_tempo(self.player)

    @tempo.setter
    def tempo(self, ms_per_quarter_note):
        """Set player tempo.

        :param ms_per_quarter_note: tempo in microseconds per beat (quarter note)
        :type ms_per_quarter_note: ``int``

        """
        return fluid_player_set_midi_tempo(self.player, ms_per_quarter_note)

    @property
    def total_ticks(self):
        """Return duration of current MIDI track in ticks.

        :rtype: ``int``

        """
        return fluid_player_get_total_ticks(self.player)


class Player(BasePlayer):
    reverb_presets = {
        # room size (0.0-1.2), damping (0.0-1.0), width (0.0-100.0), level (0.0-1.0)
        'Preset 1': (0.2, 0.0, 0.5, 0.9),
        'Preset 2': (0.4, 0.2, 0.5, 0.8),
        'Preset 3': (0.6, 0.4, 0.5, 0.7),
        'Preset 4': (0.8, 0.7, 0.5, 0.6),
        'Preset 5': (0.8, 0.0, 0.5, 0.5)
    }

    def play(self, offset=0):
        """Start playing at given offset.

        :param offset: time offset in MIDI ticks
        :type offset: ``int``

        """
        if offset != self.current_tick:
            self.seek(offset)

        return super(Player, self).play()

    def render(self, filename, filetype=None, quality=0.5, progress_callback=None):
        """Render MIDI file to audio file.

        :param filename: audio output file path and name
        'type filename: ``str``
        :param filetype: audio output file type
        :type filetype: ``str``
        :param quality: variable bit rate encoding quality when file type is
            ``flac`` or ``oga``
        :type quality: ``float`` (0.0 - 1.0)
        :param progress_callback: Python callable to call after each
            period-size block of sample frames has been written. Receives the
            filename, filetype, current total number of sample frames written
            and the period size as positional arguments in that order.
        :type progress_callback: callable with 4 positional args

        Possible choices for ``filetype`` are:

        aiff, au, auto, avr, caf, flac, htk, iff, mat, oga, paf, pvf, raw, sd2,
        sds, sf, voc, w64, wav, xi

        See also: http://www.fluidsynth.org/api/fluidsettings.xml#audio.file.type

        """
        self._set_render_settings(filename, filetype)
        renderer = new_fluid_file_renderer(self.synth.synth)
        if not renderer:
            raise OSError('Failed to create MIDI file renderer.')

        fluid_file_set_encoding_quality(renderer, quality)
        period_size = self.synth.setting('audio.period-size')
        num_samples = 0  # sample frame counter

        try:
            while self.status != FLUID_PLAYER_DONE:
                # Render one block
                if fluid_file_renderer_process_block(renderer) != FLUID_OK:
                    raise OSError('MIDI file renderer error.')

                num_samples += period_size
                if progress_callback:
                    # For progress reporting
                    progress_callback(filename, filetype, num_samples, period_size)
        finally:
            self.stop()
            self.join()
            self.synth.setting('player.timing-source', 'system')
            self.synth.setting("synth.lock-memory", 1)
            delete_fluid_file_renderer(renderer)

        return num_samples

    def _set_render_settings(self, filename, filetype=None):
        """Set audio file and audio file type and non-realtime rendering mode.

        Internal method called by ``Player.render()``.

        :param filename: audio output file path and name
        :type filename: ``str``
        :param filetype: audio output file type
        :type filetype: ``str``

        """
        if filetype is not None and filetype not in AUDIO_FILE_TYPES:
            raise OSError("Unnown file type '%s'." % filetype)

        self.synth.setting("audio.file.name", filename)
        self.synth.setting("player.timing-source", "sample")
        self.synth.setting("synth.lock-memory", 0)

        if filetype is not None:
            self.synth.setting("audio.file.type", filetype)

    def set_reverb(self, preset):
        """Change reverb preset.

        :param preset: reverb preset name (one of the ``Player.reverb_presets`` keys)
        :type preset: ``str``

        """
        self.synth.set_reverb(*self.reverb_presets[preset])

    def set_gain(self, gain):
        """Set synth master volume.

        :param gain: gain value ``0.2 - 10.0``
        :type gain: ``float``

        """
        self.synth.setting('synth.gain', float(gain))

    def stop(self):
        """Stop playing.

        Also triggers release phase of all currently sounding notes.

        """
        result = super(Player, self).stop()
        # Stop notes on all (-1) channels
        self.synth.all_notes_off(-1)
        return result


class Synth:
    """Represents a FluidSynth synthesizer."""

    def __init__(self, gain=0.2, samplerate=44100.0, channels=256, **kwargs):
        """Create new synthesizer object to control sound generation.

        Optional keyword arguments:

        :param gain: scale factor for audio output, default is 0.2. lower
            values give quieter output and allow more simultaneous notes
            without clipping
        :type gain: ``float``
        :param samplerate: output samplerate in Hz, default is 44100.0 Hz
            added capability for passing arbitrary fluid settings using args
        :type samplerate: ``float``
        :param channels: the number of MIDI channels the synthesizer supports.
            MIDI hardware is limited to 16 channels (the default value, but
            internally FluidSynth can use up to 256 channels, which can be
            mapped to different MIDI sources.
        :type channels: ``int``

        Additional keyword arguments are interpreted and applied as fluidsynth
        settings, i.e. the argument name is used as the setting name and its
        value as the setting value. Since most fluidsynth setting names are not
        valid Python identifiers and therefor can not be used keyword arguments
        directly, you have to pass them via the ``**kwarg`` syntax. The values
        must have the proper type, i.e. ``int`` for ``FLUID_INT_TYPE` settings,
        `'float`` for ``FLUID_NUM_TYPE`` and ``str``  or ``bytes`` for
        ``FLUID_STR_TYPE`` settings.

        """
        self.settings = new_fluid_settings()
        self.setting('synth.gain', float(gain))
        self.setting('synth.sample-rate', float(samplerate))
        self.setting('synth.midi-channels', channels)

        for opt, val in iteritems(kwargs):
            self.setting(opt, val)

        self.synth = new_fluid_synth(self.settings)
        self.audio_driver = None
        self.midi_driver = None
        self.router = None
        self.cmd_handler = None

    def setting(self, opt, val=None):
        """Get/Set an arbitrary synth setting, type-smart."""
        opt = _e(opt)

        if val is None:
            stype = fluid_settings_get_type(self.settings, opt)

            if stype == FLUID_NUM_TYPE:
                val = c_double()
                response = fluid_settings_getnum(self.settings, opt, byref(val))
                return val.value if response == FLUID_OK else None
            elif stype == FLUID_INT_TYPE:
                val = c_int()
                response = fluid_settings_getint(self.settings, opt, byref(val))
                return val.value if response == FLUID_OK else None
            elif stype == FLUID_STR_TYPE:
                data = create_string_buffer(256)
                response = fluid_settings_copystr(self.settings, opt, data, 256)
                return _d(data.value) if response == FLUID_OK else None
            elif stype == FLUID_SET_TYPE:
                raise NotImplementedError("Setting of type FLUID_SET_TYPE not implemented.")
            elif stype == FLUID_NO_TYPE:
                raise KeyError("Setting '%s' does not exist." % _d(opt))
        elif isinstance(val, text_type):
            return fluid_settings_setstr(self.settings, opt, _e(val))
        elif isinstance(val, binary_type):
            return fluid_settings_setstr(self.settings, opt, val)
        elif isinstance(val, bool):
            return fluid_settings_setint(self.settings, opt, 1 if val else 0)
        elif isinstance(val, int):
            return fluid_settings_setint(self.settings, opt, val)
        elif isinstance(val, float):
            return fluid_settings_setnum(self.settings, opt, val)

    def start(self, driver=None, device=None, midi_driver=None, cmd_handler=False):
        """Start audio output driver in separate background thread.

        Call this function any time after creating the Synth object to start
        handling events.

        If you don't call this function, use ``get_samples()`` to generate
        samples.

        Optional keyword arguments:

        :param driver: which audio driver to use for output
        :type driver: ``str``
        :param device: the device to use for audio output
        :type device: ``str``
        :param midi_driver: which driver to use for MIDI input
        :type midi_driver: ``str``
        :param cmd_handler: whether to create a shell command handler
        :type cmd_handler: ``bool`` (default: ``False``)

        Possible choices for ``driver`` are:

        alsa, coreaudio, dart, dsound, file, jack, oss, portaudio, pulseaudio,
        sdl2, sndman, waveout

        See also: http://www.fluidsynth.org/api/fluidsettings.xml#audio.driver

        Possible choices for ``midi_driver`` are:

        alsa_raw, alsa_seq, coremidi, jack, midishare, oss, winmidi

        See also: http://www.fluidsynth.org/api/fluidsettings.xml#midi.driver

        Not all drivers will be available for every platform, it depends on
        which drivers were compiled into FluidSynth for your platform.

        """
        if driver is not None:
            if driver not in AUDIO_DRIVER_NAMES:
                raise ValueError("Unknown audio driver '%'." % driver)
            self.setting('audio.driver', driver)

            if device is not None:
                self.setting('audio.%s.device' % driver, device)

            self.audio_driver = new_fluid_audio_driver(self.settings, self.synth)

        if midi_driver is not None:
            assert (midi_driver in ['alsa_seq','jack', 'alsa_raw', 'oss', 'winmidi', 'midishare', 'coremidi'])
            fluid_settings_setstr(self.settings, b'midi.driver', midi_driver.encode())
            if fluid_synth_set_midi_router:
                self.router = new_fluid_midi_router(self.settings, fluid_synth_handle_midi_event, self.synth)
                fluid_synth_set_midi_router(self.synth, self.router)
                self.midi_driver = new_fluid_midi_driver(self.settings, fluid_midi_router_handle_midi_event, self.router)

        if cmd_handler:
            if fluid_synth_set_midi_router:
                fluid_synth_set_midi_router(self.synth, self.router)
            else:
                self.cmd_handler = new_fluid_cmd_handler(self.synth, self.router)

    def delete(self):
        if self.audio_driver is not None:
            delete_fluid_audio_driver(self.audio_driver)

        if self.midi_driver is not None:
            delete_fluid_midi_driver(self.midi_driver)

        if self.router is not None:
            delete_fluid_midi_router(self.router)

        if self.cmd_handler is not None:
            delete_fluid_cmd_handler(self.cmd_handler)

        delete_fluid_synth(self.synth)
        delete_fluid_settings(self.settings)

    def sfload(self, filename, update_midi_preset=0):
        """Load SoundFont and return its ID."""
        if not os.path.exists(filename):
            return ValueError("No such file: "+filename)
        return fluid_synth_sfload(self.synth, _e(filename), update_midi_preset)

    def sfunload(self, sfid, update_midi_preset=0):
        """Unload a SoundFont and free memory it used."""
        return fluid_synth_sfunload(self.synth, sfid, update_midi_preset)

    def program_select(self, chan, sfid, bank, preset):
        """Select a program"""
        return fluid_synth_program_select(self.synth, chan, sfid, bank, preset)

    def program_unset(self, chan):
        """Set the preset of a MIDI channel to an unassigned state."""
        return fluid_synth_unset_program(self.synth, chan)

    def channel_info(self, chan):
        """Get soundfont, bank, prog, preset name of channel.

        Superceded by program_info and sfpreset_name, included for backwards-compatibility.

        """
        if fluid_synth_get_channel_info:
            info = fluid_synth_channel_info_t()  # noqa:F821
            fluid_synth_get_channel_info(self.synth, chan, byref(info))
            return (info.sfont_id, info.bank, info.program, info.name)

        # fluidsynth-2
        (sfontid, banknum, prognum) = self.program_info(chan)
        return (sfontid, banknum, prognum, self.sfpreset_name(sfontid, banknum, prognum))

    def program_info(self, chan):
        sfontid = c_int()
        banknum = c_int()
        prognum = c_int()
        fluid_synth_get_program(self.synth, chan, byref(sfontid), byref(banknum), byref(prognum))
        return (sfontid.value, banknum.value, prognum.value)

    def sfpreset_name(self, sfid, bank, prog):
        """Return name of a soundfont preset."""
        if not fluid_preset_get_name:
            raise NotImplementedError(
                "Fluidsynth library does not provide required 'fluid_preset_get_name' function")

        sfont = fluid_synth_get_sfont_by_id(self.synth, sfid)
        preset = fluid_sfont_get_preset(sfont, bank, prog)
        return _d(fluid_preset_get_name(preset)) if preset else None

    def router_clear(self):
        if self.router is not None:
            fluid_midi_router_clear_rules(self.router)

    def router_default(self):
        if self.router is not None:
            fluid_midi_router_set_default_rules(self.router)

    def router_add_rule(self, rule):
        if self.router is None:
            return

        if rule.rule is None:
            raise ValueError("Can't add deleted RouterRule instance.")

        return fluid_midi_router_add_rule(self.router, rule.rule, rule.type)

    def set_reverb(self, roomsize=-1.0, damping=-1.0, width=-1.0, level=-1.0):
        """Set reverb parameters.

        :param roomsize: Reverb room size value (0.0-1.0)
        :type roomsize: ``float``
        :param damping: Reverb damping value (0.0-1.0)
        :type damping: ``float``
        :param width: Reverb width value (0.0-100.0)
        :type width: ``float``
        :param level: Reverb level value (0.0-1.0)
        :type level: ``float``

        """
        if fluid_synth_set_reverb:
            return fluid_synth_set_reverb(self.synth, roomsize, damping, width, level)

        # fluidsynth-1
        mask = 0
        if roomsize >= 0:
            mask += 0b0001
        if damping >= 0:
            mask += 0b0010
        if width >= 0:
            mask += 0b0100
        if level >= 0:
            mask += 0b1000

        return fluid_synth_set_reverb_full(self.synth, mask, roomsize, damping, width, level)

    def set_chorus(self, nr=-1, level=-1.0, speed=-1.0, depth=-1.0, type=-1):
        """Set chorus parameters.

        :param nr: Chorus voice count (0-99, CPU time consumption proportional
            to this value)
        :type nr: ``int``
        :param level: Chorus level (0.0-10.0)
        :type levl: ``float``
        :param speed: Chorus speed in Hz (0.29-5.0)
        :type speed: ``float``
        :param depth: Chorus depth (max value depends on synth sample rate,
           0.0-21.0 is safe for sample rate values up to 96KHz)
        :type depth: ``float``
        :param type: Chorus waveform type (0=sine, 1=triangle)
        :type type: ``int``

        """
        if fluid_synth_set_chorus:
            return fluid_synth_set_chorus(self.synth, nr, level, speed, depth, type)

        # fluidsynth-1
        mask = 0
        if nr >= 0:
            mask += 0b00001
        if level >= 0:
            mask += 0b00010
        if speed >= 0:
            mask += 0b00100
        if depth >= 0:
            mask += 0b01000
        if type >= 0:
            mask += 0b10000

        return fluid_synth_set_chorus_full(self.synth, mask, nr, level, speed, depth, type)

    def set_reverb_roomsize(self, roomsize):
        return fluid_synth_set_reverb_roomsize(self.synth, roomsize)

    def set_reverb_damp(self, damp):
        return fluid_synth_set_reverb_damp(self.synth, damp)

    def set_reverb_level(self, level):
        return fluid_synth_set_reverb_level(self.synth, level)

    def set_reverb_width(self, width):
        return fluid_synth_set_reverb_width(self.synth, width)

    def set_chorus_nr(self, nr):
        return fluid_synth_set_chorus_nr(self.synth, nr)

    def set_chorus_level(self, level):
        return fluid_synth_set_reverb_level(self.synth, level)

    def set_chorus_speed(self, speed):
        return fluid_synth_set_chorus_speed(self.synth, speed)

    def set_chorus_depth(self, depth):
        return fluid_synth_set_chorus_depth(self.synth, depth)

    def set_chorus_type(self, type):
        return fluid_synth_set_chorus_type(self.synth, type)

    def get_reverb_roomsize(self):
        return fluid_synth_get_reverb_roomsize(self.synth)

    def get_reverb_damp(self):
        return fluid_synth_get_reverb_damp(self.synth)

    def get_reverb_level(self):
        return fluid_synth_get_reverb_level(self.synth)

    def get_reverb_width(self):
        return fluid_synth_get_reverb_width(self.synth)

    def get_chorus_nr(self):
        return fluid_synth_get_chorus_nr(self.synth)

    def get_chorus_level(self):
        return fluid_synth_get_reverb_level(self.synth)

    def get_chorus_type(self):
        return fluid_synth_get_chorus_type(self.synth)

    def get_chorus_speed(self):
        return fluid_synth_get_chorus_speed(self.synth)

    def get_chorus_depth(self):
        return fluid_synth_get_chorus_depth(self.synth)

    def noteon(self, chan, key, vel):
        """Play a note."""
        if key < 0 or key > 128:
            return False
        if chan < 0:
            return False
        if vel < 0 or vel > 128:
            return False
        return fluid_synth_noteon(self.synth, chan, key, vel)

    def noteoff(self, chan, key):
        """Stop a note."""
        if key < 0 or key > 128:
            return False

        if chan < 0:
            return False

        return fluid_synth_noteoff(self.synth, chan, key)

    def pitch_bend(self, chan, val):
        """Adjust pitch of a playing channel by small amounts.

        * A pitch bend value of 0 is no pitch change from default.
        * A value of -2048 is 1 semitone down.
        * A value of 2048 is 1 semitone up.

        Maximum values are -8192 to +8192 (transposing by 4 semitones).

        """
        return fluid_synth_pitch_bend(self.synth, chan, val + 8192)

    def cc(self, chan, ctrl, val):
        """Send control change value.

        The controls that are recognized are dependent on the
        SoundFont. Values are always 0 to 127. Typical controls
        include:

          1 : vibrato
          7 : volume
          10 : pan (left to right)
          11 : expression (soft to loud)
          64 : sustain
          91 : reverb
          93 : chorus

        """
        return fluid_synth_cc(self.synth, chan, ctrl, val)

    def get_cc(self, chan, num):
        i = c_int()
        fluid_synth_get_cc(self.synth, chan, num, byref(i))
        return i.value

    def program_change(self, chan, prg):
        """Change the program."""
        return fluid_synth_program_change(self.synth, chan, prg)

    def bank_select(self, chan, bank):
        """Choose a bank."""
        return fluid_synth_bank_select(self.synth, chan, bank)

    def sfont_select(self, chan, sfid):
        """Choose a SoundFont."""
        return fluid_synth_sfont_select(self.synth, chan, sfid)

    def program_reset(self):
        """Reset the programs on all channels."""
        return fluid_synth_program_reset(self.synth)

    def system_reset(self):
        """Stop all notes and reset all programs."""
        return fluid_synth_system_reset(self.synth)

    def all_notes_off(self, chan):
        """Turn off all notes on a MIDI channel (put them into release phase)."""
        return fluid_synth_all_notes_off(self.synth, chan)

    def get_samples(self, len=1024):
        """Generate audio samples.

        The return value will be a NumPy array containing the given
        length of audio samples. If the synth is set to stereo output
        (the default) the array will be size 2 * len.

        """
        return fluid_synth_write_s16_stereo(self.synth, len)


class Sequencer:
    def __init__(self, time_scale=1000, use_system_timer=True):
        """Create new sequencer object to control and schedule timing of midi events.

        Optional keyword arguments:

        :param time_scale: ticks per second, defaults to 1000
        :param use_system_timer: whether the sequencer should advance by itself

        """
        self.client_callbacks = []
        self.sequencer = new_fluid_sequencer2(use_system_timer)
        fluid_sequencer_set_time_scale(self.sequencer, time_scale)

    def register_fluidsynth(self, synth):
        response = fluid_sequencer_register_fluidsynth(self.sequencer, synth.synth)

        if response == FLUID_FAILED:
            raise OSError("Registering fluid synth failed")

        return response

    def register_client(self, name, callback, data=None):
        c_callback = CFUNCTYPE(None, c_uint, c_void_p, c_void_p, c_void_p)(callback)
        response = fluid_sequencer_register_client(self.sequencer, _e(name), c_callback, data)

        if response == FLUID_FAILED:
            raise OSError("Registering client failed")

        # Store in a list to prevent garbage collection
        self.client_callbacks.append(c_callback)
        return response

    def note(self, time, channel, key, velocity, duration, source=-1, dest=-1, absolute=True):
        evt = self._create_event(source, dest)
        fluid_event_note(evt, channel, key, velocity, duration)
        self._schedule_event(evt, time, absolute)
        delete_fluid_event(evt)

    def note_on(self, time, channel, key, velocity=127, source=-1, dest=-1, absolute=True):
        evt = self._create_event(source, dest)
        fluid_event_noteon(evt, channel, key, velocity)
        self._schedule_event(evt, time, absolute)
        delete_fluid_event(evt)

    def note_off(self, time, channel, key, source=-1, dest=-1, absolute=True):
        evt = self._create_event(source, dest)
        fluid_event_noteoff(evt, channel, key)
        self._schedule_event(evt, time, absolute)
        delete_fluid_event(evt)

    def timer(self, time, data=None, source=-1, dest=-1, absolute=True):
        evt = self._create_event(source, dest)
        fluid_event_timer(evt, data)
        self._schedule_event(evt, time, absolute)
        delete_fluid_event(evt)

    def _create_event(self, source=-1, dest=-1):
        evt = new_fluid_event()
        fluid_event_set_source(evt, source)
        fluid_event_set_dest(evt, dest)
        return evt

    def _schedule_event(self, evt, time, absolute=True):
        response = fluid_sequencer_send_at(self.sequencer, evt, time, absolute)

        if response == FLUID_FAILED:
            raise OSError("Scheduling event failed")

    def get_tick(self):
        return fluid_sequencer_get_tick(self.sequencer)

    def process(self, msec):
        fluid_sequencer_process(self.sequencer, msec)

    def delete(self):
        delete_fluid_sequencer(self.sequencer)



import scullery.workers
scullery.workers.start()

import scullery.iceflow
import scullery.fluidsynth

import os,time

import unittest,random,gc

class Player(scullery.iceflow.GstreamerPipeline):
    def __init__(self,file):
        scullery.iceflow.GstreamerPipeline.__init__(self,realtime=False)

        self.src = self.addElement('filesrc',location=file)

        #This bin autodetects and decodes basically any type of media
        #It is special cased, anything onnected to it is actually connected on-demand as needed
        decodebin = self.addElement('decodebin')


        self.addElement('audioconvert',connectToOutput=decodebin)
        self.addElement('audioresample')

        self.fader = self.addElement('volume', volume=1)
        self.sink = self.addElement('autoaudiosink') 


class TestAudio(unittest.TestCase):

    def test_z_no_segfaults(self):
        #Test for segfault-ery
        for i in range(0,100):
            p=Player(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "media","Brothers Unite.ogg"))
            p.start()
            time.sleep(0.01*random.random())
            p.seek(0.3)
            time.sleep(0.01*random.random())
            p.setProperty(p.fader, "volume",1)
            p.stop()
            #Ensure nothing bad happens setting the volume after stopping
            p.setProperty(p.fader, "volume",1)
        del p
        gc.collect()
        for i in range(150):
            time.sleep(0.1)
            if len(scullery.iceflow.pipes)==0:
                break
        gc.collect()

        self.assertEqual(len(scullery.iceflow.pipes), 0)
     
    def test_seekpastend(self):
        p=Player(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "media","Brothers Unite.ogg"))
        p.start()
        time.sleep(0.01*random.random())
        p.seek(99999)
        time.sleep(0.01*random.random())
        p.setProperty(p.fader, "volume",1)
        p.stop()
        #Ensure nothing bad happens setting the volume after stopping
        p.setProperty(p.fader, "volume",1)
        del p
        gc.collect()
        for i in range(150):
            time.sleep(0.1)
            if len(scullery.iceflow.pipes)==0:
                break
        gc.collect()

        self.assertEqual(len(scullery.iceflow.pipes), 0)


#Copyright Daniel Dunn 2013. 2015
#This file is part of Kaithem Automation.

#Kaithem Automation is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, version 3.

#Kaithem Automation is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Kaithem Automation.  If not, see <http://www.gnu.org/licenses/>.
import sys,os,weakref,threading,gzip,bz2,json, time,logging,hashlib,threading

posix_rename = False
if sys.platform.startswith('linux'):
    posix_rename =True
if sys.platform.startswith('darwin'):
    posix_rename =True


def resolvePath(fn,expand=True):
    if not expand:
        return fn
    return (os.path.expandvars(os.path.expanduser(fn)))



#TODO: Ensure only one thread can save a file at a time

if sys.version_info <(3,0):
    import StringIO
    strio = StringIO.StringIO
else:
    import io
    strio = io.BytesIO

persisters = []

#must be Rlock because load is recursive
lock = threading.RLock()

def saveExecutor(f):
    "Replace this with something that can call f with no args"
    f()

def saveAllAtExit():
    while persisters:
        i = persisters.pop()()
        try:
            i.save()
        except:
            logging.exception()

def chmod_private_try(p, execute=True):
    try:
        if execute:
            os.chmod(p,stat.S_IRUSR|stat.S_IWUSR|stat.S_IXUSR|stat.S_IRGRP|stat.S_IWGRP|stat.S_IXGRP)
        else:
            os.chmod(p,stat.S_IRUSR|stat.S_IWUSR|stat.S_IRGRP|stat.S_IWGRP)
    except Exception as e:
        raise e
def ensure_dir(f):
    d = os.path.dirname(f)
    if not os.path.exists(d):
        os.makedirs(d)

class Persister():
    def __init__(self,fn,default=None):
        self.fn= fn
        try:
            self.reload()
        except:
            self.value = default
        torm = []
        #Before we add ourselves, clear out any old persisters that are no longer needed.
        for i in persisters:
            if not i():
                torm.append(i)
        for i in torm:
            persisters.remove(i)

        persisters.append(weakref.ref(self))

    def save(self):
        save(self.value, self.fn, backup=True)

    def reload(self):
        if os.path.exists(self.fn):
            self.value = load(self.fn)

def save(data,fn, *,private=False,backup=True, expand=True, md5=False,nolog=False):
    """Save data to file. Filename must end in .json, .yaml, .txt, or .bin. Data will be encoded appropriately.
        Also supports compressed versions via filenames ending in .gz or .bz2.
        Args:
            data:
                the data to be written. if fn is a .json or .yaml, must be serializable. If filename is .txt, must be a string.
                If .bin, must be something like bytes.
            private:
                If True, file created with mode 700(Full access to root and owner but not even read to anyone else)
                If False(the default), file created with default mode
            backup:
                Setting this to true is an alias for mode="backup"
    """
    fn = resolvePath(fn, expand)
    with lock:

        #Make sure we don't overwrite a file when we create our dirs, because that behavior is undocumented in makedirs.
        x = os.path.split(fn)[0]
        already = {}

        #Safety counter to stop really wierd loops
        for i in range(64):
            if os.path.isfile(x):
                raise RuntimeError("Required intermediate directory is already present as a file, refusing to overwrite file")
            x = os.path.split(x)[0]
            #Loop prevention
            if x in already:
                break
            already[x] = True

        if not os.path.exists(os.path.dirname(fn)):
            if private:
                os.makedirs(os.path.dirname(fn), mode=0o700)
            else:
                os.makedirs(os.path.dirname(fn))

        if os.path.isdir(fn):
            raise RuntimeError("Filename is already present as a directory, refusing to overwrite directory")
        #Get base type without compression
        if fn.endswith(".gz"):
            x = fn[:-3]
        elif fn.endswith(".bz2"):
            x = fn[:-4]
        else:
            x=fn
        #Encode the data into our chosen format
        if x.endswith(".json"):
            data = json.dumps(data).encode('utf8')
        elif x.endswith(".yaml"):
            import yaml
            data = yaml.dump(data).encode('utf8')
        elif x.endswith(".txt") or x.endswith(".md") or x.endswith(".rst"):
            data = (str(data).encode('utf8'))
        elif x.endswith(".bin"):
            data = (data)
        else:
            raise ValueError('Unsupported or missing File Extension')



        #We have selected a compressed type. Compress in-memory first so we can read-before-write
        #Note that disk access is slow enough the call to  basically makes no difference in speed here if it's already imported
        if fn.endswith(".gz"):
            i = strio()
            f = gzip.GzipFile(fn,mode='wb',fileobj=i)
            f.write(data)
            f.close()
            data = i.getvalue()

        elif fn.endswith(".bz2"):
            c = bz2.BZ2Compressor()
            c.compress(data)
            data = c.flush()
            del c

        #Do a read-before-write. We don't write if we don't have to
        if os.path.exists(fn):
            with open(fn,"rb") as f:
                if f.read() == data:
                    return

        ensure_dir(os.path.split(fn)[0])

        if backup:
            tempfn = fn+str(time.time())
        else:
            tempfn = fn

        if not nolog:
            logging.debug("Writing: "+fn)
        #Actually write it
        with open(tempfn,'wb') as f:

            #In backup mode, pre truncate and flush.
            #This means that even if an error occors during writing, we will be able to tell by the
            #mtime that the tilde file is the right one.
            if backup:
                f.flush()
                os.fsync(f.fileno())

            #Chmod it before we write anything to it.
            if private:
                chmod_private_try(tempfn)
            f.write(data)
            f.flush()
            os.fsync(f.fileno())

        if backup:
            os.replace(tempfn, fn)

        if md5:
            with open(fn+ ".md5" , "w") as md5f:
                md5f.write(hashlib.md5(data).hexdigest())



def load(filename, *,expand=True):
    """Load a file. Return str if file extension is .txt, bytes on .bin, dict on .yaml or .json.

    After that may be a .bz2 or a .gz for compression.

    """
    filename = resolvePath(filename, expand)

    with lock:
        try:
             #Open the file and get the filename without the compression type attached to it.
            if filename.endswith(".gz"):
                f = gzip.GzipFile(filename,mode='rb')
                x = filename[:-3]
            elif filename.endswith(".bz2"):
                x = filename[:-4]
                f = bz2.BZ2File(filename,mode='rb')
            else:
                f = open(filename,'rb')
                x = filename

            if x.endswith(".json"):
                r=json.loads(f.read().decode('utf8'))
            elif x.endswith(".yaml"):
                import yaml
                r=yaml.load(f.read().decode('utf8'))
            elif x.endswith(".txt") or x.endswith(".md") or x.endswith(".rst"):
                r=f.read().decode('utf8')
            elif x.endswith(".bin"):
                r=f.read()
            else:
                raise ValueError('Unsupported File Extension')
        except Exception as e:
            try:
                f.close()
            except:
                pass
            raise
        try:
            f.close()
        except:
            pass

        return r



import os, subprocess, time,logging,select,threading,sys

log = logging.getLogger("scullery.netmedia")

default_dir = os.path.expanduser("~/Videos/IceFlow Cache/")

audioExtensions = ['.m4a','.ogg','.mp3','.wma','.opus','.flac','.wav']



def cleanupName(n):
    return n.replace("/","slash").replace("\\","backslash").replace(":",".").replace('"','').replace("[","").replace("]",'').replace("-","")

def _readStdout(proc, retVal=b''): 
  counter = 1024
  while counter:
    x =(select.select([proc.stdout],[],[],0.1)[0])
    if x: 
        x=proc.stdout.read(1)
        retVal+=x
        if(x):
            pass#sys.stdout.write(x.decode(errors="ignore"))
    else:
        break
    counter -=1
  return retVal

def downloadVideo(vid, directory=default_dir,format="", timeout=10):
    """Downoad a video by any valid youtube-dl specifier, 
    and return it's filename as soon as it starts. The file will be downloaded into directory.
    
    Should directory already contain a request bty that name, simply return that, with no extra
    network activity.
    """    
    if not os.path.isdir(directory):
        os.mkdir(directory)
    if not os.path.isdir(os.path.join(directory,"searches")):
        os.mkdir(os.path.join(directory,"searches"))
    
    searchfile=os.path.join(directory,"searches",format+"."+cleanupName(vid))

    if os.path.isfile(searchfile):
        with open(searchfile) as f:
            return os.path.join(directory,f.read())


    x = os.listdir(directory)
    log.info("Downloading:",vid)
    cmd = ['youtube-dl',"--no-part"]

    cmd+=["-o",os.path.join(directory,r'%(title)s.%(ext)s')]
    if format:
        cmd += ["-f",format]
    cmd.append(vid)
    
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE,stdin=subprocess.PIPE)
  
    op = b''
    for i in range(0,(int(timeout)*10)):
        op += _readStdout(p)
        time.sleep(0.1)
        for j in os.listdir(directory):
            #Look for the same vid that it's redownloading
            if j.encode("utf8") in op:
                if not os.path.isfile(searchfile):
                    with open(searchfile,"w") as f:
                        f.write(j)
                return os.path.join(directory, j)


            #Look for a file that was not there before
            if not j in x:
                for word in vid.split(" "):
                    #Basic check to ensure it's not unrelated crap
                    if word.lower() in j.lower():
                        #Wait to see if the command fails, if it does, delete the search file.
                        #So we can possibly try again
                        def checkSuccess():
                            if p.wait():
                                os.remove(searchfile)
                        t = threading.Thread(target=checkSuccess)


                        #Keep track of that exact search term, for cache purposes.
                        with open(searchfile,"w") as f:
                            f.write(j)
                        t.start()
                        return os.path.join(directory, j)


    raise RuntimeError("No video found")



import scullery.workers
scullery.workers.start()

import scullery.iceflow
import scullery.fluidsynth

import os,time

import unittest,random,gc

class Player(scullery.iceflow.GstreamerPipeline):
    def __init__(self,file):
        scullery.iceflow.GstreamerPipeline.__init__(self,realtime=False)

        self.src = self.addElement('filesrc',location=file)

        #This bin autodetects and decodes basically any type of media
        #It is special cased, anything onnected to it is actually connected on-demand as needed
        decodebin = self.addElement('decodebin')


        self.addElement('audioconvert',connectToOutput=decodebin)
        self.addElement('audioresample')

        self.fader = self.addElement('volume', volume=1)
        self.sink = self.addElement('autoaudiosink') 

class NonexistantElement(scullery.iceflow.GstreamerPipeline):
    def __init__(self):
        scullery.iceflow.GstreamerPipeline.__init__(self,realtime=False)

        self.src = self.addElement('nonexistantElementType')

        

class TestAudio(unittest.TestCase):
            
    def test_3s_player(self):
        print("An audio file should play for 3s, then start over and slightly speed up, then speed up for 3s")
        p=Player(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "media","Brothers Unite.ogg"))
        p.start()
        time.sleep(3)
        p.seek(0, rate=1.003)
        time.sleep(3)
        p.seek(0, rate=1.5)
        time.sleep(3)
        p.setProperty(p.fader, "volume",1)
        p.stop()
        print("The audio file should be stopped")
        del p

    def test_nonexistant_file(self):
        with self.assertRaises(ValueError):
            Player("ThisFileDoesNotExist")
            
    def test_nonexistant_element(self):
        with self.assertRaises(ValueError):
            NonexistantElement()
            

    def test_nonexistant_element(self):
        with self.assertRaises(ValueError):
            NonexistantElement()
            


#Copyright Daniel Dunn 2020
#This file is part of Scullery.

#Scullery is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, version 3.

#Scullery is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Scullery.  If not, see <http://www.gnu.org/licenses/>.

from . import workers

import time, threading,weakref,logging,types,traceback

from typing import Callable,Optional,Union
from threading import setprofile
from typeguard import typechecked


logger = logging.getLogger("tagpoints")
syslogger = logging.getLogger("system")

t = time.monotonic

#This is used for messing with the set of tags.
#We just accept that creating and deleting tags and claims is slow.
lock = threading.RLock()

allTags = {}

allTagsAtomic = {}

providers = {}

subscriberErrorHandlers=[]



from .unitsofmeasure import convert, unitTypes

class TagProvider():
    def mount(self, path):
        if not self.path.endswith("/"):
            self.path.append("/")
        self.path = path
        with lock:
            providers[path]= weakref.ref(self)
    
    def unmount(self):
        del providers[self.path]
    
    def __del__(self):
        with lock:
            del providers[self.path]

    def getTag(self, tagName):
        return _TagPoint(tagName)



#_ and . allowed
illegalCharsInName = "[]{}|\\<>,?-=+)(*&^%$#@!~`\n\r\t\0"
class _TagPoint(virtualresource.VirtualResource):
    """
        A Tag Point is a named object that can be chooses from a set of data sources based on priority,
        filters that data, and returns it on a push or a pull basis.

        A data source here is called a "Claim", and can either be a number or a function. The highest
        priority claim is called the active claim.

        If the claim is a function, it will be called at most once per interval, which is set by tag.interval=N
        in seconds. However the filter function is called every time the data is requested.

        If there are any subscribed functions to the tag, they will automatically be called at the tag's interval,
        with the one parameter being the tag's value. Any getter functions will be called to get the value.


        It is also a VirtualResource, and as such if you enter it into a module, then replace it,
        all claims and subscriptions carry over.

        One generally does not instantiate a tag this way, instead they use the Tag function
        which can get existing tags. This allows use of tags for cross=
    
    """
    defaultData=None
   
    @typechecked
    def __init__(self,name:str):
        global allTagsAtomic

        if name =="":
            raise ValueError("Tag with empty name")

        for i in illegalCharsInName:
            if i in name:
                raise ValueError("Illegal char in tag point name: "+i)
        virtualresource.VirtualResource.__init__(self)
        
        #Might be the number, or might be the getter function.
        #it's the current value of the active claim
        self._value = self.defaultData
        self.name = name
        #The cached actual value from the claims
        self.cachedRawClaimVal = self.defaultData
        #The cached output of processValue
        self.lastValue = self.defaultData
        self.lastGotValue = 0
        self.interval =0
        self.activeClaim =None
        self.claims = {}
        self.lock = threading.RLock()
        self.subscribers = []
        self.poller = None
       
        self.lastError = 0
        


      

        #String describing the "owner" of the tag point
        #This is not a precisely defined concept
        self.owner = ""

        #Stamp of when the tag's value was set
        self.timestamp = time.monotonic()
        self.annotation=None

        self.handler=None

  

       
        #If we should push the same value twice in a row when it comes in.
        #If false, only push changed data to subscribers.
        self.pushOnRepeats = False
        self.lastPushedValue=None
        self.onSourceChanged = None

        with lock:
            allTags[name]=weakref.ref(self)
            allTagsAtomic= allTags.copy()


        self.defaultClaim = self.claim(self.defaultData)
        
        #What permissions are needed to 
        #manually override this tag 
        self.permissions = []

        #This is where we can put a manual override
        #claim from the web UI.
        self.manualOverrideClaim = None

        self._alarms = {}


    @classmethod
    def Tag(cls,name:str):
        if not isinstance(name,str):
            raise TypeError("Name must be string")
        with lock:
            if name in allTags:
                x=allTags[name]()
                if x:
                    if not x.__class__ is cls:
                        raise TypeError("A tag of that name exists, but it is the wrong type.")
                    return x
            
            for i in sorted(providers.keys(),key =lambda p: len(p.path), reverse=True):
                if name.startswith(i):
                    return providers[i].getTag(i)

            return cls(name)


    @property
    def currentSource(self):
        return self.activeClaim[2]
  
    def filterValue(self,v):
        "Pure function that returns a cleaned up or normalized version of the value"
        return v

    def addAlarm(self,name, alarm):
        with self.lock:
            self._alarms[name]=weakref.ref(alarm)

            #Do some cleanup here
            torm = []
            for i in self._alarms:
                if self._alarms[i]()==None:
                    torm.append(i)
            for i in torm:
                del self._alarms[i]

    def removeAlarm(self,name):
        with self.lock:
            del self._alarms[name]

    def __del__(self):
        global allTagsAtomic
        with lock:
            try:
                del allTags[self.name]
                allTagsAtomic= allTags.copy()
            except:
                pass

    def __call__(self,*args,**kwargs):
        if not args:
            return self.value
        else:
            return self.setClaimVal(*args,**kwargs)



    def _managePolling(self):        
        if self.subscribers or self.handler:
            if not self.poller or not (self.interval == self.poller.interval):
                self.poller = scheduling.scheduler.scheduleRepeating(self.p, self.interval)
        else:
            self.poller.unregister()
            self.poller = None


    @typechecked
    def subscribe(self,f:Callable):
        with self.lock:
            
            if isinstance(f,types.MethodType):
                ref=weakref.WeakMethod(f)
            else:
                ref = weakref.ref(f)

            self.subscribers.append(ref)


            torm = []
            for i in self.subscribers:
                if not i():
                    torm.append(i)
            for i in torm:
                self.subscribers.remove(i)
    
    def unsubscribe(self,f):
        with self.lock:
            x = None
            for i in self.subscribers:
                if i()==f:
                    x = i
            if x:
                self.subscribers.remove(x)

    @typechecked
    def setHandler(self, f:Callable):
        self.handler=weakref.ref(f)

    def _guiPush(self, value):
        pass

    def _push(self,val,timestamp,annotation):
        """Push to subscribers. Only call under the same lock you changed value
            under. Otherwise the push might happen in the opposite order as the set, and
            subscribers would see the old data as most recent.
        """
       
        #This is not threadsafe, but I don't think it matters.
        #A few unnecessary updates shouldn't affect anything.
        if val==self.lastPushedValue:
            if not self.pushOnRepeats:
                return
        
        #Note the difference with the handler.
        #It is called synchronously, right then and there
        if self.handler:
            f=self.handler()
            if f:
                f(val, timestamp, annotation)
            else:
                self.handler=None
        self._guiPush(val)

        self.lastPushedValue = val

        for i in self.subscribers:
            f=i()
            if f:
                try:
                    f(val,timestamp,annotation)
                except:
                    logger.exception("Tag subscriber error")
                    #Return the error from whence it came to display in the proper place

                    for i in subscriberErrorHandlers:
                        try:
                            i(self, f,val)
                        except:
                            print("Failed to handle error: "+traceback.format_exc(6))
                   
            del f

    def processValue(self,value):

        """Represents the transform from the claim input to the output.
            Can be stateful, for things like digital filters.
        """
        #Functions are special valid types of value.
        #They are automatically resolved.
        if callable(value):
            value = value()

        return value
   
    @property
    def age(self):
        return time.time()-self.lastGotValue

    @property
    def value(self):
        return self._getValue()

    def pull(self):
        return self._getValue(True)

    def _getValue(self,force=False):
        #Get the numeric value of the tag. It is meant to be called under lock.

        activeClaimValue = self._value
        if not callable(activeClaimValue):
            self.lastValue= self.processValue(activeClaimValue)
        else:
            #Rate limited tag getter logic. We ignore the possibility for
            #Race conditions and assume that calling a little too often is fine, since
            #It shouldn't affect correctness
            if (time.time()-self.lastGotValue> self.interval) or force:
                #Set this flag immediately, or else a function with an error could defeat the cacheing
                #And just flood everything with errors
                self.lastGotValue = time.time()

                try:
                    #However, the actual logic IS ratelimited
                    #Note the lock is IN the try block so we don' handle errors under it and
                    #Cause bugs that way
                    with self.lock:
                        #None means no new data
                        x = activeClaimValue()
                        t = time.monotonic()
                        
                        if not x is None:
                            #Race here. Data might not always match timestamp an annotation, if we weren't under lock
                            self.timestamp = t 
                            self.annotation=None

                        self.cachedRawClaimVal= x or self.cachedRawClaimVal

                except:
                    #We treat errors as no new data.
                    logger.exception("Error getting tag value")

                    #The system logger is the one kaithem actually logs to file.
                    if self.lastError<(time.time()-(60*10)):
                        syslogger.exception("Error getting tag value. This message will only be logged every ten minutes.")
                    #If we can, try to send the exception back whence it came
                    try:
                        import newevt
                        newevt.eventByModuleName(activeClaimValue.__module__)._handle_exception()
                    except:
                        pass
            else:
                #processValue is responsible for it's own threadsafe-ness if any is needed
                self.lastValue = self.processValue(self.cachedRawClaimVal)
        return self.lastValue
    
    @value.setter
    def value(self, v):
        self.setClaimVal("default",v,time.monotonic(),"Set via value property")

    
    def handleSourceChanged(self,name):
        if self.onSourceChanged:
            try:
                self.onSourceChanged(name)
            except:
                logging.exception("Error handling changed source")

    def claim(self, value, name="default", priority=None,timestamp=None, annotation=None):
        """Adds a 'claim', a request to set the tag's value either to a literal 
            number or to a getter function.

            A tag's value is the highest priority claim that is currently
            active, or the value returned from the getter if the active claim is
            a function.
        """
        timestamp = timestamp or time.monotonic()

        if priority and priority>100:
            raise ValueError("Maximum priority is 100")

        if not callable(value):
            value=self.filterValue(value)
            
        with self.lock:
            #we're changing the value of an existing claim,
            #We need to get the claim object, which we stored by weakref
            claim=None
            try:
                ##If there's an existing claim by that name we're just going to modify it
                if name in self.claims:
                    claim= self.claims[name][3]()
                    #No priority change, set and return
                    if priority == claim.priority:
                        claim.set(value,timestamp, annotation)
                        return claim
                    priority= priority or claim.priority
            except:
                logger.exception("Probably a race condition and safe to ignore this")

            #If the weakref obj disappeared it will be None
            if claim ==None:
                priority = priority or 50
                claim = self.claimFactory(value,name,priority,timestamp,annotation)
        
            claim.value=value
            claim.timestamp = timestamp
            claim.annotation = annotation
            claim.priority = priority


            #Note  that we use the time, so that the most recent claim is
            #Always the winner in case of conflicts
            self.claims[name] = (priority, t(),name,weakref.ref(claim))

            if self.activeClaim==None or priority >= self.activeClaim[0]:
                self.activeClaim = self.claims[name]
                self.handleSourceChanged(name)

                self._value = value
                self.timestamp = timestamp
                self.annotation = annotation

            #If priority has been changed on the existing active claim
            #We need to handle it
            elif name==self.activeClaim[2]:
                self._findActiveClaim()

            self._push(self._getValue(),self.timestamp, self.annotation)           
            return claim

    def _findActiveClaim(self):
        #Defensive programming against weakrefs dissapearing
        #in some kind of race condition that leaves them in the list.
        #Basically we find the highest priority valid claim
        for i in reversed(sorted(self.claims.values())):
            x= i[3]()
            if x:
                self._value=x.value
                self.timestamp = x.timestamp
                self.annotation = x.annotation
                self.activeClaim=i
                self.handleSourceChanged(i[2])
                break


    def setClaimVal(self,claim,val,timestamp,annotation):
        "Set the value of an existing claim"
        timestamp = timestamp or time.monotonic()
        
        if not callable(val):
            val=self.filterValue(val)
        
        with self.lock:
            c=self.claims[claim]
            #If we're setting the active claim
            if c==self.activeClaim:
                upd=True
            else:
                upd=False
                #If we have the same priority as the active claim, even though we aren't active, by being set,
                #we can steal the active status, because this is now newer.
                if c.priority>=self.activeClaim.priority:
                    self.activeClaim=c
                    upd=True

            #Grab the claim obj and set it's val
            x= c[3]()
            x.value = val
            x.annotation=annotation
            if upd:
                self.timestamp = timestamp
                self._value=val
                self.annotation=annotation
                self._push(self._getValue(),timestamp,annotation)           


    #Get the specific claim object for this class
    def claimFactory(self, value,name,priority,timestamp,annotation):
        return Claim(self, value,name,priority,timestamp,annotation)

    def release(self, name):
        with self.lock:
            #Ifid lets us filter by ID, so that a claim object that has
            #Long since been overriden can't delete one with the same name
            #When it gets GCed
            if not name in self.claims:
                return
            
            if name=="default":
                raise ValueError("Cannot delete the default claim")

            if len(self.claims)==1:
                raise RuntimeError("Tags must maintain at least one claim")
            del self.claims[name]
            while self.claims:
                self.activeClaim = sorted(list(self.claims.values()),reverse=True)[0]
                o = self.activeClaim[3]()

                #Perhaps in a race condition that has dissapeared.
                #We must remove it and retry.
                if o==None:
                    del self.claims[self.activeClaim[2]]
                else:
                    self._value = o.value
                    self.timestamp = o.timestamp
                    self.annotation =o.annotation
                    break
            self._push(self.value, self.timestamp, self.annotation)


class _NumericTagPoint(_TagPoint):
    defaultData=0
    @typechecked
    def __init__(self,name:str, 
        min:Union[float,int,None]=None, 
        max:Union[float,int,None]=None):
       
        self._hi = 10**16
        self._lo = -10**16
        self._min=min
        self._max =max
        #Pipe separated list of how to display value
        self._displayUnits=None
        self._unit = None
        self.guiLock = threading.Lock()
        
        _TagPoint.__init__(self,name)

    def processValue(self,value):
        #Functions are special valid types of value.
        #They are automatically resolved.
        if callable(value):
            value = value()

        if self._min !=None:
            value= max(self._min,value)

        if self._max !=None:
            value= min(self._max,value)        
        
        return float(value)
   


    def filterValue(self,v):
        return float(v)

    def claimFactory(self, value,name,priority,timestamp,annotation):
        return NumericClaim(self,value,name,priority,timestamp,annotation)
    
    @property
    def min(self):
        return self._min
    
    @min.setter
    def min(self,v):
        self._min = v
        self.onConfigChanged()

    @property
    def max(self):
        return self._max
    
    @max.setter
    def max(self,v):
        self._max = v
        self.onConfigChanged()

    @property
    def hi(self):
        return self._hi
    
    @hi.setter
    def hi(self,v):
        self._hi = v
        self.onConfigChanged()

    @property
    def lo(self):
        return self._lo
    
    @lo.setter
    def lo(self,v):
        self._lo = v
        self.onConfigChanged()

    def onConfigChanged(self):
        pass

    def convertTo(self, unit):
        "Return the tag's current vakue converted to the given unit"
        return convert(self.value,self.unit,unit)
    
    def convertValue(self, value, unit):
        "Convert a value in the tag's native unit to the given unit"
        return convert(value,self.unit,unit)


    @property
    def unit(self):
        return self._unit

    @unit.setter
    @typechecked
    def unit(self,value:str):
        if self._unit:
            if not self._unit==value:
                if value:
                    raise ValueError("Cannot change unit of tagpoint. To override this, set to None first")
        self._unit = value
        #Maybe race condition here, but most likely nobody is setting this
        #from separate threads. TODO
        if not self._displayUnits:
            self._displayUnits=value
        self.onConfigChanged()
       

    @property
    def displayUnits(self):
        return self._displayUnits

    @displayUnits.setter
    def displayUnits(self,value):
        self._displayUnits = value
        self.onConfigChanged()


class _StringTagPoint(_TagPoint):
    defaultData=''
    @typechecked
    def __init__(self,name:str):
        _TagPoint.__init__(self,name)
    
    def processValue(self,value):
        #Functions are special valid types of value.
        #They are automatically resolved.
        if callable(value):
            value = value()
        
        return str(value)
   

    def filterValue(self,v):
        return str(v)

    def _guiPush(self, value):
        #Immediate write, don't push yet, do that in a thread because TCP can block
        self.spanWidget.write(value,push=False)
        def pushFunction():
            self.spanWidget.value=value
            if self.guiLock.acquire(timeout=1):
                try:
                    #Use the cached literal computed value, not what we were passed,
                    #Because it could have changed by the time we actually get to push
                    self.spanWidget.write(self.lastValue)
                finally:
                    self.guiLock.release()
        #Should there already be a function queued for this exact reason, we just let
        #That one do it's job
        if self.guiLock.acquire(timeout=0.001):
            try:
                workers.do(pushFunction)
            finally:
                self.guiLock.release()    
class Claim():
    "Represents a claim on a tag point's value"
    @typechecked
    def __init__(self,tag:_TagPoint, value, 
        name:str='default',priority:Union[int,float]=50,
        timestamp:Union[int,float,None]=None, annotation=None):

        self.name=name
        self.tag=tag
        self.value = value
        self.annotation=annotation
        self.timestamp = timestamp
        self.priority=priority
        
    def __del__(self):
        if self.name != 'default':
            self.tag.release(self.name)
    
    def set(self,value,timestamp=None, annotation=None):
        self.tag.setClaimVal(self.name, value,timestamp,annotation)


    def release(self):
        self.tag.release(self.name)

    def __call__(self,*args,**kwargs):
        if not args:
            raise ValueError("No arguments")
        else:
            return self.set(*args,**kwargs)
            
class NumericClaim(Claim):
    "Represents a claim on a tag point's value"
    @typechecked
    def __init__(self,tag:_TagPoint, value:Union[int,float], 
        name:str='default',priority:Union[int,float]=50,
        timestamp:Union[int,float,None]=None, annotation=None):

        Claim.__init__(self,tag,value,name,priority,timestamp,annotation)


    def setAs(self, value, unit, timestamp=None,annotation=None):
        "Convert a value in the given unit to the tag's native unit"
        self.set(convert(value,unit,self.tag.unit), timestamp, annotation)


Tag = _NumericTagPoint.Tag
ObjectTag = _TagPoint.Tag
StringTag = _StringTagPoint.Tag


#Copyright Daniel Dunn 2013
#This file is part of Kaithem Automation.

#Kaithem Automation is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, version 3.

#Kaithem Automation is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Kaithem Automation.  If not, see <http://www.gnu.org/licenses/>.

"This file manages the kaithem global message bus that is used mostly for logging but also for many other tasks."

import weakref,threading,time,os,random,traceback,logging,inspect,types,copy
from typing import Callable,Optional

from typeguard import typechecked

from . import workers
from collections import defaultdict, OrderedDict

_subscribers_list_modify_lock = threading.RLock()
cachelock = threading.RLock()
#OrderedDict doesn't seem as fast as dict. So I have a cache of the cache
parsecache = OrderedDict()
parsecachecache = {}

log =logging.getLogger("system.messagebus")
def normalize_topic(topic):
    """"Because some topics are equivalent("/foo" and "foo"), this lets us convert them to the canonical "/foo" representation.
    Note that "/foo/" is not the same as "/foo", because a trailing slash indicates a "directory"."""
    topic = topic.strip()
    if not topic.startswith('/'):
        return '/'+topic
    else:
        return topic

def _shouldReRaiseAttrErr():
    return True

def defaultErrorHandler(*a):
    print(traceback.format_exc())
    
subscriberErrorHandlers = []

def handleError(f,topic,value):
    log.exception("Message bus subscriber error")
    if hasattr(f,"messagebusWrapperFor"):
        f= f.messagebusWrapperFor
    for i in subscriberErrorHandlers:
        try:
            i(f,topic,value)
        except:
            print(traceback.format_exc())

def runFunction(f,a):
    return f(*a)

class MessageBus(object):
    def __init__(self,executor:Optional[Callable] = None):
        """You pass this a function of one argument that just calls its argument. Defaults to calling in
        same thread and ignoring errors.
        """
        if executor==None:
            def do(self, f):
                try:
                    f()
                except:
                    pass
            self.executor = do
        else:
            self.executor = executor

        self.subscribers = defaultdict(list)
        self.subscribers_immutable = {}

    @typechecked
    def subscribe(self,topic:str,callback:Callable):
        topic=normalize_topic(topic)
       
        with _subscribers_list_modify_lock:
            wrappedCallback = self.wrap_callback(callback,topic)

           
            self.subscribers[topic].append(wrappedCallback)
            self.subscribers_immutable = copy.deepcopy(self.subscribers)

    def unsubscribe(self,topic, function):
            "Unsubscribe topic from function"
            try:
                with _subscribers_list_modify_lock:
                    target = None
                    for j in self.subscribers[topic]:
                        if j.originalFunction()==function:
                            target = j
                    if target:
                        self.subscribers[topic].remove(target)
                    else:
                        raise ValueError("No such subscriber found")
            except:
                pass
            #There is a very slight chance someone will
            #Add something to topic before we delete it but after the test.
            #That would result in a canceled subscription
            #So we use this lock.
            try:
                with _subscribers_list_modify_lock:
                    if not self.subscribers[topic]:
                        self.subscribers.pop(topic)
            except AttributeError as e:
                #This try and if statement are supposed to catch nuisiance errors when shutting down.
                if _shouldReRaiseAttrErr():
                    raise e
 
            finally:
                with _subscribers_list_modify_lock:
                    self.subscribers_immutable = copy.deepcopy(self.subscribers)

    @staticmethod
    def parseTopic(topic):
        "Parse the topic string into a list of all subscriptions that could possibly match."
        global parsecache
        #Since this is a pure function(except the caching itself) we can cache it
        if topic in parsecache:
            return parsecache[topic]
        
        #Let's cache by the original version of the topic, so we don't have to convert it to the canonical
        oldtopic = topic
        topic=normalize_topic(topic)

        #A topic foo/bar/baz would go to
        #foo, foo/bar, and /foo/bar/baz
        #So we need to make a list like that
        matchingtopics = set(['/#'])
        parts = topic.split("/")
        last = ""

        #Add the exact one
        matchingtopics.add(topic)
        
        
        for i in parts:
            last += (i+'/')
            matchingtopics.add(last+"#")
        parsecache[oldtopic] = matchingtopics
        #Don't let the cache get too big.
        #Getting rid of the oldest should hopefully converge to the most used topics being cached
        if len(parsecache) > 600:
            parsecache.popitem(last=False)
        return matchingtopics

    @typechecked
    def wrap_callback(self,f:Callable,topic:str):
        """return function g that calls f with (topic,message) or just f(topic), depending
        on how many args there are.
         and if errors is true logs the error"""

        args = len(inspect.signature(f).parameters)


        #Allright, here is how this works.
        #We have to deal with the possibility that, at any time,
        #The callback will cease to exist. That, in fact, is how one unsubscribes.
        #So, we make this here closure that knows the topic, and
        #When the GC goes Om Nom Nom on the function, we get passed the weakref to it.
        #Then we get rid of the empty weakref and if that causes the entire topic
        #To have no subscribers, delete that too in case of memory leak.
        def delsubscription(weakrefobject):
            try:
                with _subscribers_list_modify_lock:
                    self.subscribers[topic].remove(weakrefobject)
            except:
                pass
            #There is a very slight chance someone will
            #Add something to topic before we delete it but after the test.
            #That would result in a canceled subscription
            #So we use this lock.
            try:
                with _subscribers_list_modify_lock:
                    if not self.subscribers[topic]:
                        self.subscribers.pop(topic)
            except AttributeError as e:
                #This try and if statement are supposed to catch nuisiance errors when shutting down.
                if _shouldReRaiseAttrErr():
                    raise e
               
            finally:
                with _subscribers_list_modify_lock:
                    self.subscribers_immutable = copy.deepcopy(self.subscribers)



        if isinstance(f,types.MethodType):
            f=weakref.WeakMethod(f,delsubscription)
        else:
            f = weakref.ref(f,delsubscription)

        #Mutable object that gets saved to the closure for keeping track of if we already loggged this
        alreadyLogged=[False]
        if args==0:
            def g(topic, message,errors,timestamp,annotation):
                try:
                    f2 = f()
                    if f2:
                        f2()
                except:
                    try:
                        if errors:
                            if not alreadyLogged[0]:
                                handleError(f2, topic,message)
                            alreadyLogged[0]=True
                    except Exception as e:
                            print("err",e)
        elif args==2:
            def g(topic, message,errors,timestamp,annotation):
                try:
                    f2 = f()
                    if f2:
                        f2(topic,message)
                except:
                    try:
                        if errors:
                            if not alreadyLogged[0]:
                                handleError(f,topic,message)
                            alreadyLogged[0]=True

                    except Exception as e:
                            print("err",e)
        elif args==4:
            def g(topic, message,errors,timestamp,annotation):
                try:
                    f2 = f()
                    if f2:
                        f2(topic,message,timestamp,annotation)
                except:
                    try:
                        if errors:
                            if not alreadyLogged[0]:
                                handleError(f,topic,message)
                            alreadyLogged[0]=True

                    except Exception as e:
                            print("err",e)
        elif args==1:
            def g(topic, message,errors,timestamp,annotation):
                try:
                    f2 = f()
                    if f2:
                        f2(message)
                except:
                    try:
                        if errors:
                            if not alreadyLogged[0]:
                                handleError(f,topic,message)
                            alreadyLogged[0]=True
                    except Exception as e:
                            print("err",e)
        else:
            raise ValueError("Invalid function signature(0,1,2, or 4 args supported, not "+str(args)+")")

        #Ref to the weakref so it's easy to check if the function we are wrapping
        #Still exists.
        g.originalFunction = f    
        return g



    def _post(self, topic,message,errors,timestamp,annotation,executor=None):

        executor = executor or self.executor

        matchingtopics = self.parseTopic(topic)
        #We can't iterate on anything that could possibly change so we make copies
        d = self.subscribers_immutable
        for i in matchingtopics:
            if i in d:
                #When we find a match, we make a copy of that subscriber list
                x = d[i][:]
                #And iterate the copy
                for f in x:
                    #we call the ref to get its refferent
                    #An error could happen in the subscriber
                    #Or a typeerror could because the weakref has been collected
                    #We ignore both of these errors and move on
                    executor(f,(topic,message,errors,timestamp,annotation))

    def postMessage(self, topic, message,errors=True, timestamp=None,annotation=None, synchronous=False):
        #Use the executor to run the post message job
        #To allow for the possibility of it running in the background as a thread
        topic=normalize_topic(topic)
        try:
            topic = str(topic)
        except Exception:
            raise TypeError("Topic must be a string or castable to a string.")

        timestamp = timestamp or time.monotonic()
        self._post(topic,message,errors,timestamp, annotation, runFunction if synchronous else None)


#Setup the default system messagebus
_bus = MessageBus(workers.do)
subscribe = _bus.subscribe
unsubscribe = _bus.unsubscribe
postMessage = _bus.postMessage



import unittest

from scullery import workers
workers.start()

from scullery import fluidsynth
import time

class TestAudio(unittest.TestCase):
            

    def test_fluidsynth_nojack(self):
        fs = fluidsynth.FluidSynth()
        print("Testing fluidsynth without JACK")
        print("You should hear a MIDI note. This test require's fluidsynth, pyFluidsynth, and it's default sf2 file")
        fs.noteOn(0,70,50)
        time.sleep(0.5)
        fs.noteOff(0,70)
        fs.setInstrument(0,"flute")
        print("You should hear a flute note")
        fs.noteOn(0,70,70)
        time.sleep(1)
        fs.noteOff(0,70)





from . thirdparty import baresipy
import weakref
from . import jack
import time


def wrapFunction(f):
    f = weakref.ref(f)

    def f2(*a, **k):
        f(*a, **k)
    return f2


class LocalBareSIP(baresipy.BareSIP):
    def __init__(self, *a, jackSink=None, jackSource=None, **k):
        super().__init__(pwd='', gateway='', *a, **k)
        self.jackSink = jackSink
        self.jackSource = jackSource

        self.useJack = jackSink or jackSource
        self.mostRecentCall = ''
        self.jackName = "baresip-01"
        self.jackOutName = "baresip"

    def call(self, number):
        self.mostRecentCall = number
        super().call(self, number)

    def handle_incoming_call(self, number):
        time.sleep(0.1)
        self.controller().onIncomingCall(number)
        self.mostRecentCall = number

    def handle_call_established(self):
        if self.useJack:
            # Assume we don't know which one is which.
            jn1=self.jackNames[0]
            jn2= self.jackNames[1]

            if self.jackSource:
                self.controller().inJackAirwire = jack.Airwire(self.jackSource, jn1)
                self.controller().inJackAirwire2 = jack.Airwire(self.jackSource,jn2)
                self.controller().inJackAirwire.connect()
                self.controller().inJackAirwire2.connect()

            if self.jackSink:
                self.controller().outJackAirwire = jack.Airwire(jn1, self.jackSink)
                self.controller().outJackAirwire2 = jack.Airwire(jn2, self.jackSink)
                self.controller().outJackAirwire2.connect()
                self.controller().outJackAirwire.connect()

            self.controller().onIncomingCall(self.mostRecentCall)

            # Undo the system connection it will try to do.
            if not self.jackSink == 'system':
                outPorts = jack.getPorts(
                    jn1+":*", is_output=True, is_audio=True)
                outPorts += jack.getPorts(jn2 +
                                          ":*", is_output=True, is_audio=True)

                inPorts = jack.getPorts(
                    "system:*", is_input=True, is_audio=True)
                for i in outPorts:
                    for j in inPorts:
                        jack.disconnect(i.name, j.name)

            if not self.jackSource == 'system':
                inPorts = jack.getPorts(
                    jn1+":*", is_input=True, is_audio=True)
                inPorts += jack.getPorts(jn2+":*",
                                         is_input=True, is_audio=True)
                outPorts = jack.getPorts(
                    "system:*", is_output=True, is_audio=True)
                for i in outPorts:
                    for j in inPorts:
                        jack.disconnect(i.name, j.name)

    def onJackEstablished(self, name):
        if self.useJack:
            self.controller().inJackAirwire = jack.Airwire(
                self.controller().jackSource, self.jackName)
            self.controller().outJackAirwire = jack.Airwire(
                self.controller().jackSink, self.jackName)


class SipUserAgent():
    def __init__(self, username, audioDriver="alsa,default", port=5060, jackSource=None, jackSink=None):
        super().__init__()
        self.agent = LocalBareSIP(username, audiodriver=audioDriver,
                                  port=port, jackSource=jackSource, jackSink=jackSink, block=False)
        self.agent.controller = weakref.ref(self)

    def __del__(self):
        self.agent.quit()

    def call(self, number):
        self.agent.call(number)
    
    def hang(self):
        self.agent.hang()

    def onIncomingCall(self, number):
        print(number)
        self.accept()
        time.sleep(8)
        self.accept()

    def accept(self):
        self.agent.accept_call()

    def close(self):
        self.agent.quit()




#Copyright Daniel Dunn 2013
#This file is part of Kaithem Automation.

#Kaithem Automation is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, version 3.

#Kaithem Automation is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Kaithem Automation.  If not, see <http://www.gnu.org/licenses/>.

#This file manages a work queue that feeds a threadpool
#Tasks will be performed on a best effort basis and errors will be caught and ignored.

import threading,sys,traceback, logging
import atexit,time,collections
import random

#I'm reserving the system log for a reasonable low-rate log.
#So this is in a separate namespace that shows up elsewhere.
logger = logging.getLogger("workers")

syslogger = logging.getLogger("system")

lastWorkersError = 0 
backgroundFunctionErrorHandlers=[]



def testLatency():
    start = time.monotonic()
    x=[0]
    def f():
        x[0]=time.monotonic()
    do(f)

    while time.monotonic()-start< 10:
        time.sleep(0.0001)
        if x[0]:
            return x[0]-start
    raise RuntimeError("No response")

spawnLock = threading.RLock()

maxWorkers = 32
minWorkers=4

shutdownWait= 60
run=True
taskQueue = collections.deque()

def inWaiting():
    return len(taskQueue)

def waitingtasks():
    "Return the number of tasks in the task queue"
    return len(taskQueue)


def handleErrorInFunction(f):
    print("Error in: " +str(f))

def stop():
    global run
    logging.info("Stopping worker threads")
    run = False

def EXIT():
    #Tell all worker threads to stop and wait for them all to finish.
    #If they aren't finished within the time limit, just exit.
    t = time.time()
    stop()
    for i in workers:
        try:
            #All threads total must be finished within the time limit
            workers[i].join(shutdownWait  - (time.time()-t) )
            #If we try to exit befoe the thread even has time to start or something
        except RuntimeError:
            pass

workers = {}
workersMutable ={}

wakeupHandles = []
wakeupHandlesMutable = []


def testIntegrity():
    with spawnLock:
        assert len(wakeupHandles)==len(workers)

lastStoppedThread = 0

def makeWorker(e,q,id,fastMode=False):
    #one worker that just pulls tasks from the queue and does them. Errors are caught and
    #We assume the tasks have their own error stuff
    def workerloop():
        global workers
        global lastStoppedThread
        global wakeupHandles

        f=None

        shouldRun=True

        lastActivity = time.monotonic()

        runningState = [True]
        handle = (e,runningState)
        with spawnLock:
            wakeupHandlesMutable.append(handle)
            wakeupHandles = wakeupHandlesMutable[:]

        while(run):
            try:
                runningState[0]=True
                #While either our direct  queue or the overflow queue has things in it we do them.
                while(len(taskQueue)):
                    try:
                        f=taskQueue.pop()
                    except:
                        f=None

                    if f:
                        try:
                            f[0](*f[1])
                            lastActivity=time.monotonic()
                        except Exception:
                            global lastWorkersError
                            try:
                                if lastWorkersError<time.monotonic()-60:
                                    syslogger.exception("Error in function. This message is ratelimited, see debug logs for full.\r\nIn "+f[0].__name__ +" from " + f[0].__module__ +"\r\n")
                                    lastWorkersError= time.monotonic()

                                logger.exception("Error in function running in thread pool "+f[0].__name__ +" from " + f[0].__module__)
                            except:
                                print("Failed to handle error: "+traceback.format_exc(6))

                            for i in backgroundFunctionErrorHandlers:
                                try:
                                    i(f)
                                except:
                                    print("Failed to handle error: "+traceback.format_exc(6))
                        finally:
                            #We do not want f staying around, if might hold references that should be GCed away immediatly
                            f=None
                #Ensure the lock is acqured so we can sto the next time
                #We try in blocking mode
                e.acquire(False)
                runningState[0]=False

                #This check happens *after* setting e.on false, so that if we
                #set it false right after they checked and put something in
                #the queue we get it next loop
                if not taskQueue:
                    #Randomize, so they don't all sync up
                    #FastMode polls at 100Hz
                    x = e.acquire(timeout=(random.random()*2) if not fastMode else 0.01)
                    runningState[0]=True

                if not x and not taskQueue:
                    if not shouldRun:
                        return

                    #Fast prelim check. 
                    if len(workers)>minWorkers:
                        #The elements of handle are never copied anywhere,
                        #Once the list is clear, we can be sure there is no further inserts, and the next round will catch almost
                        #all race conditions. Any remaining one in a million ones will be caught in 1 second
                        if lastActivity<(time.monotonic()-10):
                            with spawnLock:
                                if len(workers)>minWorkers :
                                    #Only stop one thread per 5 seconds to prevent
                                    #chattering

                                    #Also don't stop a thread when there aren't at least
                                    #2 threads that aren't busy.
                                    unbusyCount = 0
                                    for i in wakeupHandles:
                                        if not i[1][0]:
                                            unbusyCount+=1

                                    if unbusyCount>2 and lastStoppedThread< (time.monotonic()-5):
                                        lastStoppedThread=time.monotonic()
                                        shouldRun=None
                                        del workersMutable[id]
                                        workers = workersMutable.copy()
                                        wakeupHandlesMutable.remove(handle)
                                        wakeupHandles = wakeupHandlesMutable[:]
            except:
                print("Exception in worker loop: "+traceback.format_exc(6))

    return workerloop



def addWorker():
    global workers
    with spawnLock:
        q = []
        e = threading.Lock()
        e.acquire()

        id = time.time()
        #First worker always polls at 100hz 
        t = threading.Thread(target = makeWorker(e,q,id), name = "nostartstoplog.ThreadPoolWorker-"+str(id))
        workersMutable[id] = t
        t.start()
        workers = workersMutable.copy()

def do(func,args=[]):
    """Run a function in the background

    funct(function):
        A function of 0 arguments to be ran in the background in another thread immediatly,
    """

    taskQueue.append((func,args))

    for i in wakeupHandles:
        try:
            if i[0].locked():
                i[0].release()
                return
        except RuntimeError:
            pass
   

    #Sleep 1/25000th of a second for every item in the queue past the max number of threads
    #In an attempt to rate limit
    if len(taskQueue)>maxWorkers:
        time.sleep(max(0, (len(taskQueue)-maxWorkers)/25000  ))

    #No unbusy threads? It must go in the overflow queue.
    #Soft rate limit here should work a bit better than the old hard limit at keeping away
    #the deadlocks.
    #Under lock
    with spawnLock:
        if len(workers)< maxWorkers:
            addWorker()
            return

    #If we can't spawn a new thread
    #Wait a maximum of 15ms before 
    #just giving up and leaving 
    #it for when somethin
    #wakes up
    for n in range(0,25):
        if not taskQueue:
            return
        for i in wakeupHandles:
            try:
                if i[0].locked():
                    i[0].release()
                    return
            except RuntimeError:
                pass
        time.sleep(0.0005)


    
do_try=do

def start(count=8, qsize=64, shutdown_wait=60):
    global __queue, run,worker_wait,workers, maxWorkers
    run = True
    worker_wait = shutdown_wait
    
    maxWorkers = count
        
    syslogger.info("Started worker threads")




#Copyright Daniel Dunn 2019
#This file is part of Kaithem Automation.

#Kaithem Automation is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, version 3.

#Kaithem Automation is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License

import threading,time,logging,uuid,weakref,gc,uuid,traceback,os

from . import workers,messagebus

initialized = False
initlock = threading.Lock()

lock = threading.RLock()
Gst = None
lock = threading.RLock()
jackChannels = {}

def tryToAvoidSegfaults(t,v):
    if v.clientName=="system":
        stopAllJackUsers()
messagebus.subscribe("/system/jack/delport",tryToAvoidSegfaults)

pipes = weakref.WeakValueDictionary()

log = logging.getLogger("IceFlow_gst")

class PILCapture():
    def __init__(self,appsink):
        from PIL import Image

        self.img = Image
        self.appsink=appsink

    def pull(self):
        sample = self.appsink.emit('try-pull-sample', 0.1*10**9)
        if not sample:
            return None

        buf = sample.get_buffer()
        caps = sample.get_caps()
        h=caps.get_structure(0).get_value('height')
        w=caps.get_structure(0).get_value('width')

        return self.img.frombytes("RGB", (w, h), buf.extract_dup(0, buf.get_size()))



class PILSource():
    def __init__(self,appsrc,greyscale=False):
        self.appsrc=appsrc
        self.greyscale = greyscale


    def push(self,img):
        img = img.tobytes("raw","L" if self.greyscale else "rgb")
        img = Gst.Buffer.new_wrapped(img)
        self.appsrc.emit("push-buffer", img)

def link(a,b):
    unref = False
    try:
        if not a or not b:
            raise ValueError("Cannot link None")
        if isinstance(a, Gst.Pad):
            if not isinstance(b,Gst.Pad):
                b = b.get_static_pad("sink")
                unref=True
                if not b:
                    raise RuntimeError("B has no pad named sink and A is a pad") 
            if not a.link(b)==Gst.PadLinkReturn.OK:
                raise RuntimeError("Could not link: "+str(a)+str(b))

        elif not a.link(b):
            raise RuntimeError("Could not link"+str(a)+str(b))
    finally:
        if unref:
            pass#b.unref()

def stopAllJackUsers():
    #It seems best to stop everything using jack before stopping and starting the daemon.
    with lock:
        c = list(jackChannels.items())
        for i in c:
            #Sync stop, we gotta wait
            try:
                i[1]().syncStop=True
                i[1]().stop()
            except:
                log.exception("Err stopping JACK user")
        del c
        #Defensive programming against a double stop, which might be something that
        #No longer is in a state that can be touched without a segfault
        try:
            if i[0] in jackChannels:
                del jackChannels[i[1]]
        except:
            pass

def elementInfo(e):
    r=Gst.Registry.get()

def Element(n,name=None):
    e = Gst.ElementFactory.make(n,None)
    if e:
        return e
    else:
        raise ValueError("No such element exists: "+n)
loop = None

autorunMainloop = True

mainContext = None

def init():
    "On demand loading, for startup speed but also for compatibility if Gstreamer isn't there"
    global initialized, Gst,loop,mainContext
    #Quick check outside the lock
    if initialized:
        return
    with initlock:
        if not initialized:
            import gi
            gi.require_version('Gst', '1.0')
            gi.require_version('GstBase', '1.0')
            gi.require_version('Gtk', '3.0')
            from gi.repository import Gst as gst
            from gi.repository import GObject, GstBase, Gtk, GObject,GLib
            Gst = gst
            Gst.init(None)

            #mainContext = GLib.MainContext()

            #glibloop = GLib.MainLoop()

            #loop=threading.Thread(target=glibloop.run,daemon=True)
            #loop.start()
            initialized = True



def doesElementExist(n):
    n =Gst.ElementFactory.make(n)
    if n:
        pass#n.unref()
    return True


def wrfunc(f,fail_return=None):
    def f2(*a,**k):
        try:
            return f()(*a,**k)
        except:
            print(traceback.format_exc())
            return fail_return
    return f2

def makeWeakrefPoller(selfref,exitSignal):

    def pollerf():
        alreadyStarted = False
        t=0
        while selfref():
            self=selfref()
            
            self.threadStarted = True
            if not self.running:
                exitSignal.append(True)
                return
            
            if self.running:
                if t<time.monotonic()-3:
                    t=time.monotonic()
                    try:
                        with self.lock:
                            self.loopCallback()
                        
                            state = self.pipeline.get_state(1000000000)[1]
                        if not state==Gst.State.NULL:
                            self.wasEverRunning=True

                        #Set the flag if anything ever drives us into the null state
                        if self.wasEverRunning and state==Gst.State.NULL:
                            self.running=False
                            exitSignal.append(True)
                            return

                    except:
                        #Todo actually handle some errors?
                        exitSignal.append(True)
                        #After pipeline deleted, we clean up
                        if hasattr(self,'pipeline') and  hasattr(self,'bus'):
                            raise
                        else:
                            return
            with self.lock:
                msg = self.bus.timed_pop(100*1000)
                if msg:
                    try:
                        if msg.type == Gst.MessageType.ERROR:
                            self.on_error(self.bus,msg,None)
                        if msg.type == Gst.MessageType.EOS:
                            self.on_eos(self.bus,msg,None)
                        self.on_message(self.bus,msg,None)
                    except:
                        logging.exception("Err in pipeline:"+self.name)
                    finally:
                        pass

            del self

            #time.sleep(1)
    return pollerf




def getCaps(e):
    try:
        return e.caps
    except:
        return "UNKNOWN"
    e.getSinks()[0].getNegotiatedCaps()


def linkClosureMaker(self, src, dest,connectWhenAvailable,eid):
    "This has t be outside, it can't leak a ref to self's strong reference into the closure or it may leak memory"
    def linkFunction(element, pad,dummy):
        s = pad.query_caps(None).to_string()
        if isinstance(connectWhenAvailable,str):
            if not connectWhenAvailable in s:
                return

        if eid in self().waitingCallbacks:
            link(element,dest)
        del self().waitingCallbacks[eid]
    return linkFunction

class GStreamerPipeline():
    """Semi-immutable pipeline that presents a nice subclassable GST pipeline You can only add stuff to it.
    """
    def __init__(self, name=None, realtime=None, systemTime =False):
        init()
        self.exiting = False

        self.uuid = uuid.uuid4()
        name=name or "Pipeline"+str(time.monotonic())
        gc.collect()
        self.realtime = realtime
        self.lock = threading.RLock()
        self._stopped=True
        self.syncStop=False
        self.pipeline = Gst.Pipeline()
        self.threadStarted=False
        self.weakrefs = weakref.WeakValueDictionary()



        #This WeakValueDictionary is mostly for testing purposes
        pipes[id(self)]=self
        
        #Thread puts something in this so we know we exited
        self.exitSignal  =[]

        self.weakrefs[str(self.pipeline)]=self.pipeline
        if not self.pipeline:
            raise RuntimeError("Could not create pipeline")
        if not initialized:
            raise RuntimeError("Gstreamer not set up")
        self.bus = self.pipeline.get_bus()
        self._stopped=False

        self.weakrefs[str(self.bus)]=self.bus

        self.hasSignalWatch=0

        #Run in our own mainContext
        #mainContext.push_thread_default()
        try:
            self.bus.add_signal_watch()
        finally:
            pass#mainContext.pop_thread_default()


        self.hasSignalWatch = 1
        #1 is dummy user data, because some have reported segfaults if it is missing
        #Note that we keep strong refs to the functions, so they don't go away when we unregister,
        #Leading to a segfault in libffi because of a race condition
        self._onmessage = wrfunc(weakref.WeakMethod(self.on_message))
        self.pgbcobj = self.bus.connect('message',self._onmessage,1)
        
        self._oneos = wrfunc(weakref.WeakMethod(self.on_eos))
        self.pgbcobj2 = self.bus.connect("message::eos", self._oneos,1)

        self._onerror = wrfunc(weakref.WeakMethod(self.on_error))
        self.pgbcobj3 = self.bus.connect("message::error", self._onerror,1)
        self.name = name

        self.elements = []
        self.sidechainElements = []
        self.namedElements = {}

        self.elementTypesById= {}


        #Just a place to store refs
        self.waitingCallbacks= {}

        self.running = False
        self.wasEverRunning = True
        
        self.knownThreads= {}
        self.startTime = 0
        def dummy(*a,**k):
            pass
            
        if realtime:
            self._syncmessage = wrfunc(weakref.WeakMethod(self.syncMessage),fail_return=Gst.BusSyncReply.PASS)
            self.bus.set_sync_handler(self._syncmessage,0,dummy)
        self.pollthread=None

        self.lastElementType = None
        
        #If true, keep the pipeline running at the same rate as system time
        self.systemTime = systemTime
        self.targetRate = 1.0
        self.pipelineRate = 1.0



    def sendEOS(self):
        self.pipeline.send_event(Gst.Event.new_eos())    

    def loopCallback(self):
        #Meant to subclass. Gets called under the lock
        pass
    def seek(self, t=None,rate=None, _raw=False,_offset=0.008):
        "Seek the pipeline to a position in seconds, set the playback rate, or both"
        with self.lock:
            if self.exiting:
                return
            if not self.running:
                return
            
            if rate is None:
                rate=self.targetRate

           
            if not _raw:
                #Set "effective start time" so that the system clock sync keeps working.
                if not t is None:
                    t= max(t,0)
                    self.startTime = time.monotonic()-t
                self.targetRate = rate
                self.pipelineRate = rate
            self.pipeline.seek (rate, Gst.Format.TIME,
            Gst.SeekFlags.SKIP|Gst.SeekFlags.FLUSH, Gst.SeekType.NONE if t is None else Gst.SeekType.SET, max((t+_offset or 0)*10**9,0),
            Gst.SeekType.NONE, -1)
    
    def getPosition(self):
        "Returns stream position in seconds"
        with self.lock:
            ret,current = self.pipeline.query_position(Gst.Format.TIME)
            if not ret:
               raise RuntimeError(ret)
            if current <0:
                raise RuntimeError("Nonsense position: "+str(current))
            if current==Gst.CLOCK_TIME_NONE:
                raise RuntimeError("gst.CLOCK_TIME_NONE")
            return current/10**9

    @staticmethod
    def setCurrentThreadPriority(x,y):
        raise RuntimeError("Must override this to use realtime priority")

    def syncMessage(self,*arguments):
        "Synchronous message, so we can enable realtime priority on individual threads."
        #Stop the poorly performing sync messages after a while.
        #Wait till we have at least one thread though.
        try:
            if self.knownThreads and time.monotonic()-self.startTime>3:
                #This can't use the lock, we don't know what thread it might be called in.
                def noSyncHandler():
                    with self.lock:
                        if hasattr(self,'bus'):
                            self.bus.set_sync_handler(None,0,None)
                workers.do(noSyncHandler)
            if not threading.currentThread().ident in self.knownThreads:
                self.knownThreads[threading.currentThread().ident] = True
                if self.realtime:
                    try:
                        self.setCurrentThreadPriority(1,self.realtime)
                    except:
                        log.exception("Error setting realtime priority")
            return Gst.BusSyncReply.PASS
        except:
            return Gst.BusSyncReply.PASS
            print(traceback.format_exc())
    
    def makeElement(self,n,name=None):
        with self.lock:
            e = Element(n,name)
            self.elementTypesById[id(e)] = n
            self.pipeline.add(e)
            return e

    
    #Low level wrapper just for filtering out args we don't care about
    def on_eos(self,*a,**k):
        self.onEOS()

    def onEOS(self):
        def f2():
            try:
                self.stop()
            except:
                pass
        workers.do(f2)

        self.onStreamFinished()

    def onStreamFinished(self):
        pass

    def __del__(self):
        self.running=False
        t=time.monotonic()
        
        #Give it some time, in case it really was started
        if not self.threadStarted:
            time.sleep(0.01)
            time.sleep(0.01)
            time.sleep(0.01)
            time.sleep(0.01)
            
        if self.threadStarted:
            while not self.exitSignal:
                time.sleep(0.1)
                if time.monotonic()-t> 10:
                    raise RuntimeError("Timeout")
                
        with self.lock:
            if not self._stopped:
                self.stop()


       
        

    def on_message(self, bus, message,userdata):
        s = message.get_structure()
        if s:
            self.onMessage(message.src,s.get_name(), s)
        return True

    def onMessage(self,src,name, structure):
        pass

    def on_error(self,bus,msg,userdata):
        logging.debug('Error {}: {}, {}'.format(msg.src.name, *msg.parse_error()))



    def _waitForState(self,s,timeout=10):
        t=time.monotonic()
        while not self.pipeline.get_state(1000_000_000)[1]==s:
            if time.monotonic()-t> timeout:
                raise RuntimeError("Timeout, pipeline still in: ", self.pipeline.get_state(1000_000_000)[1])
            time.sleep(0.1)

    def start(self, effectiveStartTime=None,timeout=10):
        "effectiveStartTime is used to keep multiple players synced when used with systemTime"
        with lock:
            with self.lock:
                if self.exiting:
                    return

                            
                x = effectiveStartTime or time.time()
                timeAgo = time.time()-x
                #Convert to monotonic time that the nternal APIs use
                self.startTime= time.monotonic()-timeAgo
                
                
                if not self.pipeline.get_state(1000_000_000)[1] ==Gst.State.PAUSED:
                    self.pipeline.set_state(Gst.State.PAUSED)
                    self._waitForState(Gst.State.PAUSED)
                
                #Seek to where we should be, if we had actually
                #Started when we should have. We want to get everything set up in the pause state
                #First so we have the right "effective" start time.

                #We accept cutting off a few 100 milliseconds if it means
                #staying synced.
                if self.systemTime:
                    self.seek(time.monotonic()-self.startTime)

                self.pipeline.set_state(Gst.State.PLAYING)
                self._waitForState(Gst.State.PLAYING,timeout)
                self.running=True

                for i in range(0,500):
                    try:
                        #Test that we can actually read the clock
                        self.getPosition()
                        break
                    except:
                        if i>150:
                            raise RuntimeError("Clock still not valid")
                        time.sleep(0.1)

                #Don't start the thread until we have a valid clock
                self.maybeStartPoller()

    def play(self):
        with self.lock:
            if self.exiting:
                return
            if not self.running:
                raise RuntimeError("Pipeline is not paused, or running, call start()")
            if not self.pipeline.get_state(1000_000_000)[1] in (Gst.State.PLAYING,Gst.State.PAUSED,Gst.State.READY):
                raise RuntimeError("Pipeline is not paused, or running, call start()")
            
            #Hopefully this willl raise an error if the clock is invalid for some reason,
            #Instead of potentially causing a segfault, if that was the problem
            self.getPosition()
            
            self.pipeline.set_state(Gst.State.PLAYING)
            self._waitForState(Gst.State.PLAYING)

    def pause(self):
        "Not that we can start directly into paused without playing first, to preload stuff"
        with self.lock:
            if self.exiting:
                return
            self.pipeline.set_state(Gst.State.PAUSED)
            self._waitForState(Gst.State.PAUSED)
            self.getPosition()
            self.running=True
            self.maybeStartPoller()
    
    def maybeStartPoller(self):
        if not self.pollthread:
            self.pollthread = threading.Thread(target=makeWeakrefPoller(weakref.ref(self),self.exitSignal),daemon=True,name="nostartstoplog.GSTPoller")
            self.pollthread.daemon=True
            self.pollthread.start()

    def stop(self):
        with lock:
            #Actually stop as soon as we can
            with self.lock:
                if hasattr(self,'pipeline'):
                    #This was causing segfaults for some reasons
                    if not (self.pipeline.get_state(1000_000_000)[1]==Gst.State.NULL):
                        self.pipeline.set_state(Gst.State.NULL)
                self.exiting = True
                if hasattr(self,'bus'):
                    self.bus.set_sync_handler(None,0,None)
                    if self.hasSignalWatch:
                        self.bus.remove_signal_watch()
                        self.hasSignalWatch = False
        
        #Now we're going to do the cleanup stuff
        #In the background, because it involves a lot of waiting.
        #This might fail, if it never even started, but we just kinda ignore that.
        def gstStopCleanupTask():
            self.running=False
            t = time.monotonic()
            time.sleep(0.01)
            
            
            if not self.threadStarted:
                time.sleep(0.01)
                time.sleep(0.01)
                time.sleep(0.01)
                time.sleep(0.01)
            
            #On account of the race condition, it is possible that the thread actually never did start yet
            #So we have to ignore the exit flag stuff.
            
            #It shouldn't really be critical, most likely the thread can stop on it's own time anyway, because it doesn't do anything without getting the lock.
            if self.threadStarted:
                while not self.exitSignal:
                    time.sleep(0.1)
                    if time.monotonic()-t> 10:
                        break
                
            with self.lock:
                if self._stopped:
                    return

                self.pipeline.set_state(Gst.State.NULL)
                self._waitForState(Gst.State.NULL,10000)

                #This stuff happens in the NULL state, because we prefer not to mess with stuff while it's
                #Running
                try:
                    self.bus.disconnect(self.pgbcobj)
                    del self.pgbcobj
                    self.bus.disconnect(self.pgbcobj2)
                    del self.pgbcobj2
                    self.bus.disconnect(self.pgbcobj3)
                    del self.pgbcobj3
                except:
                    print(traceback.format_exc())
    

                del self.elements
                del self.namedElements
                del self.pipeline
                del self.bus

                try:
                    del jackChannels[self.uuid]
                except:
                    pass
                self._stopped=True

        #Allow waiting so we can be real sure it's stopped
        if not self.syncStop:
            workers.do(gstStopCleanupTask)
        else:
            gstStopCleanupTask()
        
       

    @staticmethod
    def shouldAllowGstJack():
        raise RuntimeError("You must override this method to return True if jackd is running and can be used")

    def addPILCapture(self,resolution,connectToOutput=None, buffer=1):
        "Return a video capture object"
        conv = self.addElement("videoconvert",connectToOutput=connectToOutput)
        scale=self.addElement("videoscale")
        caps = self.addElement("capsfilter",caps="video/x-raw,width="+str(resolution[0])+",height="+str(resolution[0])+", format=RGB")
        appsink = self.addElement("appsink",drop=True,sync=False,max_buffers=buffer)

        return PILCapture(appsink)

    def addPILSource(self,resolution, buffer=1,greyscale=False):
        "Return a video source object that we can use to put PIL buffers into the stream"

        appsrc = self.addElement("appsrc",caps="video/x-raw,width="+str(resolution[0])+",height="+str(resolution[0])+", format="+"GREy8" if greyscale else "RGB",connectToOutput=False)
        conv = self.addElement("videoconvert")
        scale=self.addElement("videoscale")

        #Start with a blck image to make things prerooll
        if(greyscale):
            appsrc.emit("push-buffer", Gst.Buffer.new_wrapped(bytes(resolution[0]*resolution[1])))
        else:
            appsrc.emit("push-buffer",Gst.Buffer.new_wrapped( bytes(resolution[0]*resolution[1]*3)))

      

        return PILSource(appsrc,greyscale)
    
    def addElement(self,t,name=None,connectWhenAvailable=False, connectToOutput=None, sidechain=False, **kwargs):

        #Don't let the user use JACK if it's not running,
        #For fear of gstreamer undefined behavior
        if t.startswith("jackaudio"):
            if not self.shouldAllowGstJack():
                raise RuntimeError("JACK not running")
        if lock.acquire(timeout=10):
            try:
                with self.lock:
                    if not isinstance(t, str):
                        raise ValueError("Element type must be string")

                    e = Gst.ElementFactory.make(t,name)
                    
                    if e==None:
                        raise ValueError("Nonexistant element type: "+t)
                    self.weakrefs[str(e)]=e
                    self.elementTypesById[id(e)] = t


                    for i in kwargs:
                        v = kwargs[i]
                        self.setProperty(e,i,v)
                        
                    self.pipeline.add(e)

                
                    if connectToOutput:
                        if not id(connectToOutput) in self.elementTypesById:
                            raise ValueError("Cannot connect to the output of: "+str(connectToOutput)+", no such element in pipeline.")
                    
                    #Element doesn't have an input pad, we want this to be usable as a fake source to go after a real source if someone
                    #wants to use it as a effect
                    if t=="audiotestsrc":
                        connectToOutput=False
                        


                    #This could be the first element
                    if self.elements and (not (connectToOutput is False)):
                        connectToOutput=connectToOutput or self.elements[-1]

                        #Fakesinks have no output, we automatically don't connect those
                        if self.elementTypesById[id(connectToOutput)]=='fakesink':
                            connectToOutput=False

                        #Decodebin doesn't have a pad yet for some awful reason
                        elif (self.elementTypesById[id(connectToOutput)]=='decodebin') or connectWhenAvailable:
                            eid = uuid.uuid4()
                            f = linkClosureMaker(weakref.ref(self),connectToOutput,e,connectWhenAvailable, eid)

                            self.waitingCallbacks[eid]=f
                            #Dummy 1 param because some have claimed to get segfaults without
                            connectToOutput.connect("pad-added",f,1)
                        else:
                            link(connectToOutput,e)
                    
                    #Sidechain means don't set this element as the
                    #automatic thing that the next entry links to
                    if not sidechain:
                        self.elements.append(e)
                    else:
                        self.sidechainElements.append(e)

                    self.namedElements[name]=e

                    #Mark as a JACK user so we can stop if needed for JACK
                    #Stuff
                    if t.startswith("jackaudio"):
                        with lock:
                            jackChannels[self.uuid] = weakref.ref(self)

                    self.lastElementType = t
                    p= weakref.proxy(e)
                    #List it under the proxy as well
                    self.elementTypesById[id(p)] = t
                    return p

            finally:
                lock.release()
        else:
            raise RuntimeError("Getting lock")
        
    def setProperty(self, element, prop,value):
        with self.lock:

            if prop=="location" and self.elementTypesById[id(element)]=='filesrc':
                if not os.path.isfile(value):
                    raise ValueError("No such file: "+value)
                
            if prop=='caps':
                value=Gst.Caps(value)
                self.weakrefs[str(value)]=value

            prop=prop.replace("_","-")

            prop=prop.split(":")
            if len(prop)>1:
                childIndex=int(prop[0])
                target= element.get_child_by_index(childIndex)
                target.set_property(prop[1], value)
                self.weakrefs[str(target)+"fromgetter"]=target
            else:
                element.set_property(prop[0], value)
    def isActive(self):
        with self.lock:
            if self.pipeline.get_state(1000_000_000)[1] ==Gst.State.PAUSED:
                return True
            if self.pipeline.get_state(1000_000_000)[1] ==Gst.State.PLAYING:
                return True
#Legacy misspelling
GstreamerPipeline= GStreamerPipeline